"""Flows for downloading data from Azure SQL and uploading it to Azure ADLS."""

from typing import Any

from prefect import flow
from prefect.task_runners import ConcurrentTaskRunner

from viadot.orchestration.prefect.tasks import azure_sql_to_df, df_to_adls


@flow(
    name="Azure SQL extraction to ADLS",
    description="Extract data from Azure SQL"
    + " and load it into Azure Data Lake Storage.",
    retries=1,
    retry_delay_seconds=60,
    task_runner=ConcurrentTaskRunner,
    log_prints=True,
)
def azure_sql_to_adls(
    query: str | None = None,
    credentials_secret: str | None = None,
    validate_df_dict: dict[str, Any] | None = None,
    convert_bytes: bool = False,
    remove_special_characters: bool | None = None,
    columns_to_clean: list[str] | None = None,
    adls_config_key: str | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    r"""Download data from Azure SQL to a CSV file and uploading it to ADLS.

    Args:
        query (str): Query to perform on a database. Defaults to None.
        credentials_secret (str, optional): The name of the Azure Key Vault
            secret containing a dictionary with database credentials.
            Defaults to None.
        validate_df_dict (Dict[str], optional): A dictionary with optional list of
            tests to verify the output dataframe. If defined, triggers the `validate_df`
            task from task_utils. Defaults to None.
        convert_bytes (bool). A boolean value to trigger method df_converts_bytes_to_int
            It is used to convert bytes data type into int, as pulling data with bytes
            can lead to malformed data in data frame.
            Defaults to False.
        remove_special_characters (str, optional): Call a function that remove
            special characters like escape symbols. Defaults to None.
        columns_to_clean (List(str), optional): Select columns to clean, used with
            remove_special_characters. If None whole data frame will be processed.
            Defaults to None.
        adls_config_key (Optional[str], optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        adls_azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret containing a dictionary with ACCOUNT_NAME and Service Principal
            credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
        adls_path (Optional[str], optional): Azure Data Lake destination file path (with
            file name). Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to True.
    """
    data_frame = azure_sql_to_df(
        query=query,
        credentials_secret=credentials_secret,
        validate_df_dict=validate_df_dict,
        convert_bytes=convert_bytes,
        remove_special_characters=remove_special_characters,
        columns_to_clean=columns_to_clean,
    )

    return df_to_adls(
        df=data_frame,
        path=adls_path,
        credentials_secret=adls_azure_key_vault_secret,
        config_key=adls_config_key,
        overwrite=adls_path_overwrite,
    )



"""Flows for pulling data from/into Sharepoint."""

import contextlib
from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_databricks, sharepoint_to_df


@flow(
    name="extract--sharepoint--databricks",
    description="Extract data from Sharepoint and load it into Databricks.",
    retries=1,
    retry_delay_seconds=60,
)
def sharepoint_to_databricks(
    sharepoint_url: str,
    databricks_table: str,
    databricks_schema: str | None = None,
    if_exists: Literal["replace", "skip", "fail"] = "fail",
    sheet_name: str | list | int | None = None,
    columns: str | list[str] | list[int] | None = None,
    sharepoint_credentials_secret: str | None = None,
    sharepoint_config_key: str | None = None,
    databricks_credentials_secret: str | None = None,
    databricks_config_key: str | None = None,
) -> None:
    """Download a file from Sharepoint and upload it to Azure Data Lake.

    Args:
        sharepoint_url (str): The URL to the file.
        databricks_table (str): The name of the target table.
        databricks_schema (str, optional): The name of the target schema.
        if_exists (str, Optional): What to do if the table already exists.
            One of 'replace', 'skip', and 'fail'.
        columns (str | list[str] | list[int], optional): Which columns to ingest.
            Defaults to None.
        sheet_name (str | list | int, optional): Strings are used for sheet names.
            Integers are used in zero-indexed sheet positions
            (chart sheets do not count as a sheet position). Lists of strings/integers
            are used to request multiple sheets. Specify None to get all worksheets.
            Defaults to None.
        sharepoint_credentials_secret (str, optional): The name of the Azure Key Vault
            secret storing relevant credentials. Defaults to None.
        sharepoint_config_key (str, optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        databricks_credentials_secret (str, optional): The name of the Azure Key Vault
            secret storing relevant credentials. Defaults to None.
        databricks_config_key (str, optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
    """
    # Workaround Prefect converting this parameter to string due to multiple
    # supported input types -- pandas relies on the data type to choose relevant
    # implementation.
    if sheet_name is not None:
        with contextlib.suppress(ValueError):
            sheet_name = int(sheet_name)

    df = sharepoint_to_df(
        url=sharepoint_url,
        credentials_secret=sharepoint_credentials_secret,
        config_key=sharepoint_config_key,
        sheet_name=sheet_name,
        columns=columns,
    )
    return df_to_databricks(
        df=df,
        schema=databricks_schema,
        table=databricks_table,
        if_exists=if_exists,
        credentials_secret=databricks_credentials_secret,
        config_key=databricks_config_key,
    )



from viadot.orchestration.prefect.flows import (
    cloud_for_customers_to_adls,
    cloud_for_customers_to_databricks,
)
from viadot.sources import AzureDataLake, Databricks


TEST_SCHEMA = "test_viadot_schema"
TEST_TABLE = "test"


def test_cloud_for_customers_to_adls(
    cloud_for_customers_url,
    TEST_FILE_PATH,
    c4c_credentials_secret,
    adls_credentials_secret,
):
    lake = AzureDataLake(config_key="adls_test")

    assert not lake.exists(TEST_FILE_PATH)

    cloud_for_customers_to_adls(
        cloud_for_customers_url=cloud_for_customers_url,
        adls_path=TEST_FILE_PATH,
        cloud_for_customers_credentials_secret=c4c_credentials_secret,
        adls_credentials_secret=adls_credentials_secret,
    )

    assert lake.exists(TEST_FILE_PATH)

    lake.rm(TEST_FILE_PATH)


def test_cloud_for_customers_to_databricks(
    cloud_for_customers_url, c4c_credentials_secret, databricks_credentials_secret
):
    databricks = Databricks(config_key="databricks-qa-elt")

    assert not databricks._check_if_table_exists(schema=TEST_SCHEMA, table=TEST_TABLE)

    cloud_for_customers_to_databricks(
        cloud_for_customers_url=cloud_for_customers_url,
        databricks_table=TEST_TABLE,
        databricks_schema=TEST_SCHEMA,
        cloud_for_customers_credentials_secret=c4c_credentials_secret,
        databricks_credentials_secret=databricks_credentials_secret,
    )

    assert databricks._check_if_table_exists(schema=TEST_SCHEMA, table=TEST_TABLE)

    databricks.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE)

    databricks.session.stop()



"""'mediatool_to_adls.py'."""

from typing import Any

from prefect import flow
from prefect.task_runners import ConcurrentTaskRunner

from viadot.orchestration.prefect.tasks import df_to_adls, mediatool_to_df


@flow(
    name="Mediatool extraction to ADLS",
    description="Extract data from Mediatool and load it into Azure Data Lake Storage.",
    retries=1,
    retry_delay_seconds=60,
    task_runner=ConcurrentTaskRunner,
)
def mediatool_to_adls(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    organization_ids: list[str] | None = None,
    media_entries_columns: list[str] | None = None,
    adls_credentials: dict[str, Any] | None = None,
    adls_config_key: str | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    """Download data from Mediatool to Azure Data Lake.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (str, optional): The name of the Azure Key Vault secret
            where credentials are stored. Defaults to None.
        organization_ids (list[str], optional): List of organization IDs.
            Defaults to None.
        media_entries_columns (list[str], optional): Columns to get from media entries.
            Defaults to None.
        adls_credentials (dict[str, Any], optional): The credentials as a dictionary.
            Defaults to None.
        adls_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        adls_azure_key_vault_secret (str, optional): The name of the Azure Key Vault
            secret containing a dictionary with ACCOUNT_NAME and Service Principal
            credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
        adls_path (str, optional): Azure Data Lake destination file path.
            Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to True.
    """
    data_frame = mediatool_to_df(
        config_key=config_key,
        azure_key_vault_secret=azure_key_vault_secret,
        organization_ids=organization_ids,
        media_entries_columns=media_entries_columns,
    )

    return df_to_adls(
        df=data_frame,
        path=adls_path,
        credentials=adls_credentials,
        credentials_secret=adls_azure_key_vault_secret,
        config_key=adls_config_key,
        overwrite=adls_path_overwrite,
    )



"""Genesys Cloud API connector."""

import asyncio
import base64
from io import StringIO
import json
import time
from typing import Any

import aiohttp
from aiolimiter import AsyncLimiter
import numpy as np
import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.exceptions import APIError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, handle_api_response, validate


class GenesysCredentials(BaseModel):
    """Validate Genesys credentials.

    Two key values are held in the Genesys connector:
        - client_id: The unique ID for the organization.
        - client_secret: Secret string of characters to have access to divisions.

    Args:
        BaseModel (pydantic.main.ModelMetaclass): A base class for creating Pydantic
            models.
    """

    client_id: str
    client_secret: str


class Genesys(Source):
    ENVIRONMENTS = (
        "cac1.pure.cloud",
        "sae1.pure.cloud",
        "mypurecloud.com",
        "usw2.pure.cloud",
        "aps1.pure.cloud",
        "apne3.pure.cloud",
        "apne2.pure.cloud",
        "mypurecloud.com.au",
        "mypurecloud.jp",
        "mypurecloud.ie",
        "mypurecloud.de",
        "euw2.pure.cloud",
        "euc2.pure.cloud",
        "mec1.pure.cloud",
    )

    def __init__(
        self,
        *args,
        credentials: GenesysCredentials | None = None,
        config_key: str = "genesys",
        verbose: bool = False,
        environment: str = "mypurecloud.de",
        **kwargs,
    ):
        """Genesys Cloud API connector.

        Provides functionalities for connecting to Genesys Cloud API and downloading
        generated reports. It includes the following features:

        - Generate reports inside Genesys.
        - Download the reports previously created.
        - Direct connection to Genesys Cloud API, via GET method, to retrieve the data
          without any report creation.
        - Remove any report previously created.

        Args:
            credentials (Optional[GenesysCredentials], optional): Genesys credentials.
                Defaults to None
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to "genesys".
            verbose (bool, optional): Increase the details of the logs printed on the
                screen. Defaults to False.
            environment (str, optional): the domain that appears for Genesys Cloud
                Environment based on the location of your Genesys Cloud organization.
                Defaults to "mypurecloud.de".

        Examples:
            genesys = Genesys(
                credentials=credentials,
                config_key=config_key,
                verbose=verbose,
                environment=environment,
            )
            genesys.api_connection(
                endpoint=endpoint,
                queues_ids=queues_ids,
                view_type=view_type,
                view_type_time_sleep=view_type_time_sleep,
                post_data_list=post_data_list,
                normalization_sep=normalization_sep,
            )
            data_frame = genesys.to_df(
                drop_duplicates=drop_duplicates,
                validate_df_dict=validate_df_dict,
            )

        Raises:
            CredentialError: If credentials are not provided in local_config or directly
                as a parameter.
            APIError: When the environment variable is not among the available.
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(GenesysCredentials(**raw_creds))

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.verbose = verbose
        self.data_returned = {}
        self.new_report = "{}"  # ???

        if environment in self.ENVIRONMENTS:
            self.environment = environment
        else:
            raise APIError(
                f"Environment '{environment}' not available"
                + " in Genesys Cloud Environments."
            )

    @property
    def headers(self) -> dict[str, Any]:
        """Get request headers.

        Returns:
            Dict[str, Any]: Request headers with token.
        """
        client_id = self.credentials.get("client_id", "")
        client_secret = self.credentials.get("client_secret", "")
        authorization = base64.b64encode(
            bytes(client_id + ":" + client_secret, "ISO-8859-1")
        ).decode("ascii")
        request_headers = {
            "Authorization": f"Basic {authorization}",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        request_body = {"grant_type": "client_credentials"}
        response = handle_api_response(
            f"https://login.{self.environment}/oauth/token",
            data=request_body,
            headers=request_headers,
            method="POST",
            timeout=3600,
        )

        if response.ok:
            self.logger.info("Temporary authorization token was generated.")
        else:
            self.logger.info(
                f"Failure: { response.status_code !s} - { response.reason }"
            )
        response_json = response.json()

        return {
            "Authorization": f"{ response_json['token_type'] }"
            + f" { response_json['access_token']}",
            "Content-Type": "application/json",
        }

    def _api_call(
        self,
        endpoint: str,
        post_data_list: list[str],
        method: str,
        params: dict[str, Any] | None = None,
        time_between_api_call: float = 0.5,
    ) -> dict[str, Any]:
        """General method to connect to Genesys Cloud API and generate the response.

        Args:
            endpoint (str): Final end point to the API.
            post_data_list (List[str]): List of string templates to generate json body.
            method (str): Type of connection to the API. Defaults to "POST".
            params (Optional[Dict[str, Any]], optional): Parameters to be passed into
                the POST call. Defaults to None.
            time_between_api_call (int, optional): The time, in seconds, to sleep the
                call to the API. Defaults to 0.5.

        Raises:
            RuntimeError: There is no current event loop in asyncio thread.

        Returns:
            Dict[str, Any]: Genesys Cloud API response. When the endpoint requires to
                create a report within Genesys Cloud, the response is just useless
                information. The useful data must be downloaded from apps.{environment}
                through another requests.
        """
        limiter = AsyncLimiter(2, 15)
        semaphore = asyncio.Semaphore(value=1)
        url = f"https://api.{self.environment}/api/v2/{endpoint}"

        async def generate_post():
            for data_to_post in post_data_list:
                payload = json.dumps(data_to_post)

                async with aiohttp.ClientSession() as session:
                    await semaphore.acquire()

                    async with limiter:
                        if method == "POST":
                            async with session.post(
                                url,
                                headers=self.headers,
                                data=payload,
                            ) as resp:
                                # global new_report
                                self.new_report = await resp.read()
                                message = "Generated report export ---"
                                if self.verbose:
                                    message += f"\n {payload}."
                                    self.logger.info(message)

                                semaphore.release()

                        elif method == "GET":
                            async with session.get(
                                url,
                                headers=self.headers,
                                params=params,
                            ) as resp:
                                self.new_report = await resp.read()
                                message = "Connecting to Genesys Cloud"
                                if self.verbose:
                                    message += f": {params}."
                                    self.logger.info(message)

                                semaphore.release()

                await asyncio.sleep(time_between_api_call)

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError as err:
            if str(err).startswith("There is no current event loop in thread"):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            else:
                raise
        coroutine = generate_post()
        loop.run_until_complete(coroutine)

        return json.loads(self.new_report.decode("utf-8"))

    def _load_reporting_exports(
        self,
        page_size: int = 100,
    ) -> dict[str, Any]:
        """Consult the status of the reports created in Genesys Cloud.

        Args:
            page_size (int, optional): The number of items on page to print.
                Defaults to 100.
            verbose (bool, optional): Switch on/off for logging messages.
                Defaults to False.

        Raises:
            APIError: Failed to loaded the exports from Genesys Cloud.

        Returns:
            Dict[str, Any]: Schedule genesys report.
        """
        response = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/"
            + f"analytics/reporting/exports?pageSize={page_size}",
            headers=self.headers,
            method="GET",
        )

        response_ok = 200
        if response.status_code == response_ok:
            return response.json()

        self.logger.error(f"Failed to loaded all exports. - {response.content}")
        msg = "Failed to loaded all exports."
        raise APIError(msg)

    def _get_reporting_exports_url(self, entities: list[str]) -> tuple[list[str]]:
        """Collect all reports created in Genesys Cloud.

        Args:
            entities (List[str]): List of dictionaries with all the reports information
                available in Genesys Cloud.

        Returns:
            Tuple[List[str]]: A tuple with Lists of IDs and URLs.
        """
        ids = []
        urls = []
        status = []
        for entity in entities:
            ids.append(entity.get("id"))
            urls.append(entity.get("downloadUrl"))
            status.append(entity.get("status"))

        if "FAILED" in status:
            self.logger.error("Some reports have not been successfully created.")
        if "RUNNING" in status:
            self.logger.warning(
                "Some reports are still being created and can not be downloaded."
            )
        if self.verbose:
            message = "".join(
                [f"\t{i} -> {j} \n" for i, j in zip(ids, status, strict=False)]
            )
            self.logger.info(f"Report status:\n{message}")

        return ids, urls

    def _delete_report(self, report_id: str) -> None:
        """Delete a particular report in Genesys Cloud.

        Args:
            report_id (str): Id of the report to be deleted.
        """
        delete_response = handle_api_response(
            url=f"https://api.{self.environment}/api/v2/"
            + f"analytics/reporting/exports/{report_id}",
            headers=self.headers,
            method="DELETE",
        )
        # Ok-ish responses (includes eg. 204 No Content)
        ok_response_limit = 300
        if delete_response.status_code < ok_response_limit:
            self.logger.info(
                f"Successfully deleted report '{report_id}' from Genesys API."
            )
        else:
            self.logger.error(
                f"Failed to delete report '{report_id}' "
                + f"from Genesys API. - {delete_response.content}"
            )

    def _download_report(
        self,
        report_url: str,
        drop_duplicates: bool = True,
    ) -> pd.DataFrame:
        """Download report from Genesys Cloud.

        Args:
            report_url (str): url to report, fetched from json response.
            drop_duplicates (bool, optional): Decide if drop duplicates.
                Defaults to True.

        Returns:
            pd.DataFrame: Data in a pandas DataFrame.
        """
        response = handle_api_response(url=f"{report_url}", headers=self.headers)

        # Ok-ish responses (includes eg. 204 No Content)
        ok_response_limit = 300
        if response.status_code < ok_response_limit:
            self.logger.info(
                f"Successfully downloaded report from Genesys API ('{report_url}')."
            )

        else:
            msg = (
                "Failed to download report from"
                + f" Genesys API ('{report_url}'). - {response.content}"
            )
            self.logger.error(msg)

        dataframe = pd.read_csv(StringIO(response.content.decode("utf-8")))

        if drop_duplicates is True:
            dataframe.drop_duplicates(inplace=True, ignore_index=True)

        return dataframe

    def _merge_conversations(self, data_to_merge: list) -> pd.DataFrame:  # noqa: C901, PLR0912
        """Merge all the conversations data into a single data frame.

        Args:
            data_to_merge (list): List with all the conversations in json format.
            Example for all levels data to merge:
                {
                "conversations": [
                    {
                        **** LEVEL 0 data ****
                        "participants": [
                            {
                                **** LEVEL 1 data ****
                                "sessions": [
                                    {
                                        "agentBullseyeRing": 1,
                                        **** LEVEL 2 data ****
                                        "mediaEndpointStats": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                        "metrics": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                        "segments": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                    }
                                ],
                            },
                            {
                                "participantId": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                                **** LEVEL 1 data ****
                                "sessions": [
                                    {
                                        **** LEVEL 2 data ****
                                        "mediaEndpointStats": [
                                            {
                                                **** LEVEL 3 data ****
                                            }
                                        ],
                                        "flow": {
                                            **** LEVEL 2 data ****
                                        },
                                        "metrics": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                        "segments": [
                                            {
                                                **** LEVEL 3 data ****
                                            },
                                        ],
                                    }
                                ],
                            },
                        ],
                    }
                ],
                "totalHits": 100,
            }

        Returns:
            DataFrame: A single data frame with all the content.
        """
        # LEVEL 0
        df0 = pd.json_normalize(data_to_merge)
        df0.drop(["participants"], axis=1, inplace=True)

        # LEVEL 1
        df1 = pd.json_normalize(
            data_to_merge,
            record_path=["participants"],
            meta=["conversationId"],
        )
        df1.drop(["sessions"], axis=1, inplace=True)

        # LEVEL 2
        df2 = pd.json_normalize(
            data_to_merge,
            record_path=["participants", "sessions"],
            meta=[
                ["participants", "externalContactId"],
                ["participants", "participantId"],
            ],
            errors="ignore",
            sep="_",
        )
        # Columns that will be the reference for the next LEVEL
        df2.rename(
            columns={
                "participants_externalContactId": "externalContactId",
                "participants_participantId": "participantId",
            },
            inplace=True,
        )
        for key in ["metrics", "segments", "mediaEndpointStats"]:
            try:
                df2.drop([key], axis=1, inplace=True)
            except KeyError as err:
                self.logger.info(f"Key {err} not appearing in the response.")

        # LEVEL 3
        conversations_df = {}
        for i, conversation in enumerate(data_to_merge):
            # Not all "sessions" have the same data, and that creates
            #   problems of standardization
            # Empty data will be added to columns where there is not to avoid
            #   future errors.
            for j, entry_0 in enumerate(conversation["participants"]):
                for key in list(entry_0.keys()):
                    if key == "sessions":
                        for k, entry_1 in enumerate(entry_0[key]):
                            if "metrics" not in list(entry_1.keys()):
                                conversation["participants"][j][key][k]["metrics"] = []
                            if "segments" not in list(entry_1.keys()):
                                conversation["participants"][j][key][k]["segments"] = []
                            if "mediaEndpointStats" not in list(entry_1.keys()):
                                conversation["participants"][j][key][k][
                                    "mediaEndpointStats"
                                ] = []

            # LEVEL 3 metrics
            df3_1 = pd.json_normalize(
                conversation,
                record_path=["participants", "sessions", "metrics"],
                meta=[
                    ["participants", "sessions", "sessionId"],
                ],
                errors="ignore",
                record_prefix="metrics_",
                sep="_",
            )
            df3_1.rename(
                columns={"participants_sessions_sessionId": "sessionId"}, inplace=True
            )

            # LEVEL 3 segments
            df3_2 = pd.json_normalize(
                conversation,
                record_path=["participants", "sessions", "segments"],
                meta=[
                    ["participants", "sessions", "sessionId"],
                ],
                errors="ignore",
                record_prefix="segments_",
                sep="_",
            )
            df3_2.rename(
                columns={"participants_sessions_sessionId": "sessionId"}, inplace=True
            )

            # LEVEL 3 mediaEndpointStats
            df3_3 = pd.json_normalize(
                conversation,
                record_path=["participants", "sessions", "mediaEndpointStats"],
                meta=[
                    ["participants", "sessions", "sessionId"],
                ],
                errors="ignore",
                record_prefix="mediaEndpointStats_",
                sep="_",
            )
            df3_3.rename(
                columns={"participants_sessions_sessionId": "sessionId"}, inplace=True
            )

            # merging all LEVELs 3 from the same conversation
            dff3_tmp = pd.concat([df3_1, df3_2])
            dff3 = pd.concat([dff3_tmp, df3_3])

            conversations_df.update({i: dff3})

        # MERGING ALL LEVELS
        # LEVELS 3
        for i_3, key in enumerate(list(conversations_df.keys())):
            if i_3 == 0:
                dff3_f = conversations_df[key]
            else:
                dff3_f = pd.concat([dff3_f, conversations_df[key]])

        # LEVEL 3 with LEVEL 2
        dff2 = pd.merge(dff3_f, df2, how="outer", on=["sessionId"])

        # LEVEL 2 with LEVEL 1
        dff1 = pd.merge(
            df1, dff2, how="outer", on=["externalContactId", "participantId"]
        )

        # LEVEL 1 with LEVEL 0
        return pd.merge(df0, dff1, how="outer", on=["conversationId"])

    # This is way too complicated for what it's doing...
    def api_connection(  # noqa: PLR0912, PLR0915, C901.
        self,
        endpoint: str | None = None,
        queues_ids: list[str] | None = None,
        view_type: str | None = None,
        view_type_time_sleep: int = 10,
        post_data_list: list[dict[str, Any]] | None = None,
        time_between_api_call: float = 0.5,
        normalization_sep: str = ".",
    ) -> None:
        """General method to connect to Genesys Cloud API and generate the response.

        Args:
            endpoint (Optional[str], optional): Final end point to the API.
                Defaults to None.

                Custom endpoints have specific key words, and parameters:
                Example:
                    - "routing/queues/{id}/members": "routing_queues_members"
                    - members_ids = ["xxxxxxxxx", "xxxxxxxxx", ...]
            queues_ids (Optional[List[str]], optional): List of queues ids to consult
                the members. Defaults to None.
            view_type (Optional[str], optional): The type of view export job to be
                created. Defaults to None.
            view_type_time_sleep (int, optional): Waiting time to retrieve data from
                Genesys Cloud API. Defaults to 10.
            post_data_list (Optional[List[Dict[str, Any]]], optional): List of string
                templates to generate json body in POST calls to the API.
                Defaults to None.
            time_between_api_call (int, optional): The time, in seconds, to sleep the
                call to the API. Defaults to 0.5.
            normalization_sep (str, optional): Nested records will generate names
                separated by sep. Defaults to ".".

        Raises:
            APIError: Some or No reports were not created.
            APIError: At different endpoints:
                - 'analytics/conversations/details/query': only one body must be used.
                - 'routing_queues_members': extra parameter `queues_ids` must be
                    included.
        """
        self.logger.info(
            f"Connecting to the Genesys Cloud using the endpoint: {endpoint}"
        )

        if endpoint == "analytics/reporting/exports":
            self._api_call(
                endpoint=endpoint,
                post_data_list=post_data_list,
                time_between_api_call=time_between_api_call,
                method="POST",
            )

            msg = (
                f"Waiting {view_type_time_sleep} seconds for"
                + " caching data from Genesys Cloud API."
            )
            self.logger.info(msg)
            time.sleep(view_type_time_sleep)

            request_json = self._load_reporting_exports()
            entities = request_json["entities"]

            if isinstance(entities, list):
                ids, urls = self._get_reporting_exports_url(entities)
                if len(entities) != len(post_data_list):
                    self.logger.warning(
                        f"There are {len(entities)} available reports in Genesys, "
                        f"and where sent {len(post_data_list)} reports. "
                        "Unsed reports will be removed."
                    )
            else:
                APIError(
                    "There are no reports to be downloaded."
                    f"May be {view_type_time_sleep} should be increased."
                )

            # download and delete reports created
            count = 0
            raise_api_error = False
            for qid, url in zip(ids, urls, strict=False):
                if url is not None:
                    df_downloaded = self._download_report(report_url=url)

                    time.sleep(1.0)
                    # remove resume rows
                    if view_type in ["queue_performance_detail_view"]:
                        criteria = (
                            df_downloaded["Queue Id"]
                            .apply(lambda x: str(x).split(";"))
                            .apply(lambda x: not len(x) > 1)
                        )
                        df_downloaded = df_downloaded[criteria]

                    self.data_returned.update({count: df_downloaded})
                else:
                    self.logger.error(
                        f"Report id {qid} didn't have time to be created. "
                        + "Consider increasing the `view_type_time_sleep` parameter "
                        + f">> {view_type_time_sleep} seconds to allow Genesys Cloud "
                        + "to conclude the report creation."
                    )
                    raise_api_error = True

                self._delete_report(qid)

                count += 1  # noqa: SIM113

            if raise_api_error:
                msg = "Some reports creation failed."
                raise APIError(msg)

        elif endpoint == "analytics/conversations/details/query":
            if len(post_data_list) > 1:
                msg = "Not available more than one body for this end-point."
                raise APIError(msg)

            stop_loop = False
            page_counter = post_data_list[0]["paging"]["pageNumber"]
            self.logger.info(
                "Restructuring the response in order to be able to insert it into a "
                + "data frame.\n\tThis task could take a few minutes.\n"
            )
            while not stop_loop:
                report = self._api_call(
                    endpoint=endpoint,
                    post_data_list=post_data_list,
                    time_between_api_call=time_between_api_call,
                    method="POST",
                )

                merged_data_frame = self._merge_conversations(report["conversations"])
                self.data_returned.update(
                    {
                        int(post_data_list[0]["paging"]["pageNumber"])
                        - 1: merged_data_frame
                    }
                )

                if page_counter == 1:
                    max_calls = int(np.ceil(report["totalHits"] / 100))
                if page_counter == max_calls:
                    stop_loop = True

                post_data_list[0]["paging"]["pageNumber"] += 1
                page_counter += 1

        elif endpoint in ["routing/queues", "users"]:
            page = 1
            self.logger.info(
                "Restructuring the response in order to be able to insert it into a "
                + "data frame.\n\tThis task could take a few minutes.\n"
            )
            while True:
                if endpoint == "routing/queues":
                    params = {"pageSize": 500, "pageNumber": page}
                elif endpoint == "users":
                    params = {
                        "pageSize": 500,
                        "pageNumber": page,
                        "expand": "presence,dateLastLogin,groups"
                        + ",employerInfo,lasttokenissued",
                        "state": "any",
                    }
                response = self._api_call(
                    endpoint=endpoint,
                    post_data_list=post_data_list,
                    time_between_api_call=time_between_api_call,
                    method="GET",
                    params=params,
                )

                if response["entities"]:
                    df_response = pd.json_normalize(
                        response["entities"],
                        sep=normalization_sep,
                    )
                    self.data_returned.update({page - 1: df_response})

                    page += 1
                else:
                    break

        elif endpoint == "routing_queues_members":
            counter = 0
            if queues_ids is None:
                self.logger.error(
                    "This endpoint requires `queues_ids` parameter to work."
                )
                APIError("This endpoint requires `queues_ids` parameter to work.")

            for qid in queues_ids:
                self.logger.info(f"Downloading Agents information from Queue: {qid}")
                page = 1
                while True:
                    response = self._api_call(
                        endpoint=f"routing/queues/{qid}/members",
                        params={"pageSize": 100, "pageNumber": page},
                        post_data_list=post_data_list,
                        time_between_api_call=time_between_api_call,
                        method="GET",
                    )

                    if response["entities"]:
                        df_response = pd.json_normalize(response["entities"])
                        # drop personal information
                        columns_to_drop = {
                            "user.addresses",
                            "user.primaryContactInfo",
                            "user.images",
                        }.intersection(df_response.columns)
                        df_response.drop(
                            columns_to_drop,
                            axis=1,
                            inplace=True,
                        )
                        self.data_returned.update({counter: df_response})

                        page += 1
                        counter += 1
                    else:
                        break

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: str = "warn",
        **kwargs,
    ) -> pd.DataFrame:
        """Generate a pandas DataFrame from self.data_returned.

        Args:
            drop_duplicates (bool, optional): Remove duplicates from the DataFrame.
                Defaults to False.
            validate_df_dict (Optional[Dict[str, Any]], optional): A dictionary with
                optional list of tests to verify the output dataframe. Defaults to None.

        Returns:
            pd.Dataframe: The response data as a pandas DataFrame plus viadot metadata.
        """
        drop_duplicates = kwargs.get("drop_duplicates", False)
        validate_df_dict = kwargs.get("validate_df_dict", None)
        super().to_df(if_empty=if_empty)

        for key in list(self.data_returned.keys()):
            if key == 0:
                data_frame = self.data_returned[key]
            else:
                data_frame = pd.concat([data_frame, self.data_returned[key]])

        if drop_duplicates:
            data_frame.drop_duplicates(inplace=True)

        if validate_df_dict:
            validate(df=data_frame, tests=validate_df_dict)

        if len(self.data_returned) == 0:
            data_frame = pd.DataFrame()
            self._handle_if_empty(
                if_empty=if_empty,
                message="The response does not contain any data.",
            )
        else:
            data_frame.reset_index(inplace=True, drop=True)

        return data_frame



"""Flows for pulling data from/into Sharepoint."""

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_adls, sharepoint_to_df


@flow(
    name="extract--sharepoint--adls",
    description="Extract data from Exchange Rates API and load it into Azure Data Lake.",
    retries=1,
    retry_delay_seconds=60,
)
def sharepoint_to_adls(
    sharepoint_url: str,
    adls_path: str,
    sharepoint_credentials_secret: str | None = None,
    sharepoint_config_key: str | None = None,
    adls_credentials_secret: str | None = None,
    adls_config_key: str | None = None,
    sheet_name: str | list[str | int] | int | None = None,
    columns: str | list[str] | list[int] | None = None,
    overwrite: bool = False,
) -> None:
    """Download a file from Sharepoint and upload it to Azure Data Lake.

    Args:
        sharepoint_url (str): The URL to the file.
        adls_path (str): The destination path.
        sharepoint_credentials_secret (str, optional): The name of the Azure Key Vault
            secret storing the Sharepoint credentials. Defaults to None.
        sharepoint_config_key (str, optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        adls_credentials_secret (str, optional): The name of the Azure Key Vault secret
            storing the ADLS credentials. Defaults to None.
        adls_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        sheet_name (str | list | int, optional): Strings are used
            for sheet names. Integers are used in zero-indexed sheet positions
            (chart sheets do not count as a sheet position). Lists of strings/integers
            are used to request multiple sheets. Specify None to get all worksheets.
            Defaults to None.
        columns (str | list[str] | list[int], optional): Which columns to ingest.
            Defaults to None.
        overwrite (bool, optional): Whether to overwrite files in the lake. Defaults
            to False.
    """
    df = sharepoint_to_df(
        url=sharepoint_url,
        credentials_secret=sharepoint_credentials_secret,
        config_key=sharepoint_config_key,
        sheet_name=sheet_name,
        columns=columns,
    )
    return df_to_adls(
        df=df,
        path=adls_path,
        credentials_secret=adls_credentials_secret,
        config_key=adls_config_key,
        overwrite=overwrite,
    )



from viadot.orchestration.prefect.tasks import bcp


SCHEMA = "sandbox"
TABLE = "test_bcp"
ERROR_TABLE = "test_bcp_error"
ERROR_LOG_FILE = "log_file.log"
TEST_CSV_FILE_PATH = "test_bcp.csv"


def test_bcp():
    try:
        result = bcp(
            credentials_secret="sql-server",  # noqa: S106
            path=TEST_CSV_FILE_PATH,
            schema=SCHEMA,
            table=TABLE,
            error_log_file_path=ERROR_LOG_FILE,
        )
    except Exception:
        result = False
    assert result is not False



"""'test_customer_gauge.py'."""

from unittest.mock import MagicMock, patch

import pytest

from viadot.exceptions import APIError
from viadot.sources.customer_gauge import CustomerGauge, CustomerGaugeCredentials


variables = {
    "credentials": {
        "client_id": "client_id",
        "client_secret": "client_secret",  # pragma: allowlist secret
    }
}


@pytest.mark.basic
def test_customer_gauge_credentials():
    """Test Customer Gauge credentials."""
    CustomerGaugeCredentials(
        client_id=variables["credentials"]["client_id"],
        client_secret=variables["credentials"]["client_secret"],
    )


@pytest.mark.basic
def test_missing_credentials():
    """Test raise error without Customer Gauge credentials."""
    with pytest.raises(TypeError):
        CustomerGauge(config_key="invalid_key")


@pytest.mark.connect
@patch("viadot.sources.customer_gauge.handle_api_response")
def test_get_token_success(mock_handle_api_response):
    """Test Customer Gauge get token."""
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": "fake_token"}
    mock_handle_api_response.return_value = mock_response

    customer_gauge_instance = CustomerGauge(credentials=variables["credentials"])

    token = customer_gauge_instance._get_token()
    assert token == "fake_token"  # noqa: S105 # pragma: allowlist secret
    mock_handle_api_response.assert_called_once()


@pytest.mark.connect
@patch("viadot.sources.customer_gauge.handle_api_response")
def test_get_token_failure(mock_handle_api_response):
    """Test Customer Gauge get token failure."""
    mock_response = MagicMock()
    mock_response.json.return_value = {"access_token": None}
    mock_handle_api_response.return_value = mock_response

    customer_gauge_instance = CustomerGauge(credentials=variables["credentials"])

    with pytest.raises(
        APIError, match="The token could not be generated. Check your credentials."
    ):
        customer_gauge_instance._get_token()


@pytest.mark.connect
@patch(
    "viadot.sources.customer_gauge.CustomerGauge._get_token", return_value="fake_token"
)
@patch("viadot.sources.customer_gauge.handle_api_response")
def test_get_json_response(mock_handle_api_response, mock_get_token):  # noqa: ARG001
    """Test Customer Gauge json response."""
    mock_response = MagicMock()
    mock_response.json.return_value = {"data": [], "cursor": {"next": 1}}
    mock_handle_api_response.return_value = mock_response

    customer_gauge_instance = CustomerGauge(credentials=variables["credentials"])

    json_response = customer_gauge_instance._get_json_response(url="fake_url")
    assert "data" in json_response


@pytest.mark.connect
def test_get_cursor():
    """Test Customer Gauge get cursor."""
    json_response = {"cursor": {"next": 10}}
    customer_gauge_instance = CustomerGauge(credentials=variables["credentials"])

    cursor = customer_gauge_instance._get_cursor(json_response)
    assert cursor == 10


@pytest.mark.connect
def test_get_cursor_key_error():
    """Test Customer Gauge get cursor error."""
    json_response = {"no_cursor": {}}
    customer_gauge_instance = CustomerGauge(credentials=variables["credentials"])

    with pytest.raises(
        ValueError, match="Provided argument doesn't contain 'cursor' value."
    ):
        customer_gauge_instance._get_cursor(json_response)


@pytest.mark.functions
def test_column_unpacker():
    """Test Customer Gauge function `_unpack_columns`."""
    json_list = [{"field": {"key1": "value1", "key2": "value2"}}]
    customer_gauge_instance = CustomerGauge(credentials=variables["credentials"])

    unpacked_list = customer_gauge_instance._unpack_columns(
        json_list=json_list, unpack_by_field_reference_cols=["field"]
    )
    assert "key1" in unpacked_list[0]["field"]



"""Flows for downloading data from Sharepoint and uploading it to AWS Redshift Spectrum."""  # noqa: W505

from typing import Any, Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import (
    df_to_redshift_spectrum,
    sharepoint_to_df,
)


@flow(
    name="extract--sharepoint--redshift_spectrum",
    description="Extract data from Sharepoint and load it into AWS Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
def sharepoint_to_redshift_spectrum(  # noqa: PLR0913
    sharepoint_url: str,
    to_path: str,
    schema_name: str,
    table: str,
    tests: dict[str, Any] | None = None,
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: str | None = None,
    sep: str = ",",
    aws_config_key: str | None = None,
    credentials_secret: str | None = None,
    sheet_name: str | list[str | int] | int | None = None,
    columns: str | list[str] | list[int] | None = None,
    na_values: list[str] | None = None,
    sharepoint_credentials_secret: str | None = None,
    sharepoint_config_key: str | None = None,
    file_sheet_mapping: dict | None = None,
) -> None:
    """Extract data from SharePoint and load it into AWS Redshift Spectrum.

    This function downloads data either from SharePoint file or the whole directory and
    uploads it to AWS Redshift Spectrum.

    Modes:
    If the `URL` ends with the file (e.g ../file.xlsx) it downloads only the file and
    creates a table from it.
    If the `URL` ends with the folder (e.g ../folder_name/): it downloads multiple files
    and creates a table from them:
        - If `file_sheet_mapping` is provided, it downloads and processes only
            the specified files and sheets.
        - If `file_sheet_mapping` is NOT provided, it downloads and processes all of
            the files from the chosen folder.


    Args:
        sharepoint_url (str): The URL to the file.
        to_path (str): Path to a S3 folder where the table will be located. Defaults to
            None.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        tests (dict[str], optional): A dictionary with optional list of tests
            to verify the output dataframe. If defined, triggers the `validate`
            function from viadot.utils. Defaults to None.
        partition_cols (list[str]): List of column names that will be used to create
            partitions. Only takes effect if dataset=True.
        extension (str): Required file type. Accepted file formats are 'csv' and
            'parquet'.
        if_exists (str, optional): 'overwrite' to recreate any possible existing table
            or 'append' to keep any possible existing table. Defaults to overwrite.
        partition_cols (list[str], optional): List of column names that will be used to
            create partitions. Only takes effect if dataset=True. Defaults to None.
        index (bool, optional): Write row names (index). Defaults to False.
        compression (str, optional): Compression style (None, snappy, gzip, zstd).
        sep (str, optional): Field delimiter for the output file. Defaults to ','.
        aws_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        credentials_secret (str, optional): The name of a secret block in Prefect
            that stores AWS credentials. Defaults to None.
        sheet_name (str | list | int, optional): Strings are used for sheet names.
            Integers are used in zero-indexed sheet positions (chart sheets do not count
            as a sheet position). Lists of strings/integers are used to request multiple
            sheets. Specify None to get all worksheets. Defaults to None.
        columns (str | list[str] | list[int], optional): Which columns to ingest.
            Defaults to None.
        na_values (list[str] | None): Additional strings to recognize as NA/NaN.
            If list passed, the specific NA values for each column will be recognized.
            Defaults to None.
        sharepoint_credentials_secret (str, optional): The name of the secret storing
            Sharepoint credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        sharepoint_config_key (str, optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        file_sheet_mapping (dict): A dictionary where keys are filenames and values are
            the sheet names to be loaded from each file. If provided, only these files
            and sheets will be downloaded. Defaults to None.
    """
    df = sharepoint_to_df(
        url=sharepoint_url,
        sheet_name=sheet_name,
        tests=tests,
        columns=columns,
        na_values=na_values,
        file_sheet_mapping=file_sheet_mapping,
        credentials_secret=sharepoint_credentials_secret,
        config_key=sharepoint_config_key,
    )
    df_to_redshift_spectrum(
        df=df,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        index=index,
        compression=compression,
        sep=sep,
        config_key=aws_config_key,
        credentials_secret=credentials_secret,
    )



"""Flows for downloading data from SQLServer and uploading it to MinIO."""

from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_minio, sql_server_to_df


@flow(
    name="extract--sql_server--minio",
    description="Extract data from SQLServer and load it into MinIO.",
    retries=1,
    retry_delay_seconds=60,
)
def sql_server_to_minio(
    query: str,
    path: str,
    if_exists: Literal["error", "delete_matching", "overwrite_or_ignore"] = "error",
    basename_template: str | None = None,
    sql_server_credentials_secret: str | None = None,
    sql_server_config_key: str | None = None,
    minio_credentials_secret: str | None = None,
    minio_config_key: str | None = None,
) -> None:
    """Download a file from SQLServer and upload it to MinIO.

    Args:
        query (str, required): The query to execute on the SQL Server database.
            If the query doesn't start with "SELECT" returns an empty DataFrame.
        path (str): Path to the MinIO file/folder.
        basename_template (str, optional): A template string used to generate
            base names of written data files. The token '{i}' will be replaced with
            an automatically incremented integer. Defaults to None.
        if_exists (Literal["error", "delete_matching", "overwrite_or_ignore"],
            optional). What to do if the dataset already exists. Defaults to "error".
        sql_server_credentials_secret (str, optional): The name of the secret storing
            the credentials to the SQLServer. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        sql_server_config_key (str, optional): The key in the viadot config holding
            relevant credentials to the SQLServer. Defaults to None.
        minio_credentials_secret (str, optional): The name of the secret storing
            the credentials to the MinIO. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        minio_config_key (str, optional): The key in the viadot config holding relevant
            credentials to the MinIO. Defaults to None.
    """
    df = sql_server_to_df(
        query=query,
        config_key=sql_server_config_key,
        credentials_secret=sql_server_credentials_secret,
    )

    return df_to_minio(
        df=df,
        path=path,
        if_exists=if_exists,
        basename_template=basename_template,
        config_key=minio_config_key,
        credentials_secret=minio_credentials_secret,
    )



"""Tasks for interacting with SQLServer."""

from typing import Any, Literal

import pandas as pd
from prefect import task
from prefect.logging import get_run_logger

from viadot.config import get_source_credentials
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources.base import Record
from viadot.sources.sql_server import SQLServer


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60 * 3)
def create_sql_server_table(
    schema: str,
    table: str,
    if_exists: Literal["fail", "replace", "skip", "delete"] = "fail",
    dtypes: dict[str, Any] | None = None,
    credentials_secret: str | None = None,
    config_key: str | None = None,
) -> None:
    """A task for creating table in SQL Server.

    Args:
        schema (str): Destination schema.
        table (str): Destination table.
        if_exists (Literal, optional): What to do if the table already exists.
        dtypes (dict[str, Any], optional): Data types to enforce.
        credentials_secret (str, optional): The name of the secret storing
            the credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = get_credentials(credentials_secret) or get_source_credentials(
        config_key
    )
    sql_server = SQLServer(credentials=credentials)

    fqn = f"{schema}.{table}" if schema is not None else table
    created = sql_server.create_table(
        schema=schema, table=table, dtypes=dtypes, if_exists=if_exists
    )
    if created:
        logger.info(f"Successfully created table {fqn}.")
    else:
        logger.info(
            f"Table {fqn} has not been created as if_exists is set to {if_exists}."
        )


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60 * 3)
def sql_server_to_df(
    query: str,
    credentials_secret: str | None = None,
    config_key: str | None = None,
) -> pd.DataFrame:
    """Load the result of a SQL Server Database query into a pandas DataFrame.

    Args:
        query (str, required): The query to execute on the SQL Server database.
            If the query doesn't start with "SELECT" returns an empty DataFrame.
        credentials_secret (str, optional): The name of the secret storing
            the credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.

    Returns: pd.Dataframe
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = get_source_credentials(config_key) or get_credentials(
        credentials_secret
    )
    sql_server = SQLServer(credentials=credentials)
    df = sql_server.to_df(query=query)
    nrows = df.shape[0]
    ncols = df.shape[1]

    logger.info(
        f"Successfully downloaded {nrows} rows and {ncols} columns of data to a DataFrame."
    )
    return df


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60 * 3)
def sql_server_query(
    query: str,
    credentials_secret: str | None = None,
    config_key: str | None = None,
) -> list[Record] | bool:
    """Execute a query on SQL Server.

    Args:
        query (str, required): The query to execute on the SQL Server database.
        credentials_secret (str, optional): The name of the secret storing
            the credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = get_source_credentials(config_key) or get_credentials(
        credentials_secret
    )
    sql_server = SQLServer(credentials=credentials)
    result = sql_server.run(query)

    logger.info("Successfully ran the query.")
    return result



"""Tests for the SharepointList class."""

from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
import requests
from sharepy.errors import AuthError

from viadot.exceptions import CredentialError
from viadot.sources import SharepointList
from viadot.sources.sharepoint import SharepointCredentials


DUMMY_CREDS = {
    "site": "tenant.sharepoint.com",  # pragma: allowlist secret`
    "username": "user@example.com",  # pragma: allowlist secret`
    "password": "password",  # pragma: allowlist secret`
}

LIST_NAME = "my_list"
LIST_SITE = "my_site"


@pytest.fixture
def mock_response():
    """Create a reusable mock response object."""
    return MagicMock()


@pytest.fixture
def mock_connection():
    """Create a reusable mock connection object."""
    conn = MagicMock()
    conn.site = "test.sharepoint.com"
    conn.get = MagicMock()
    return conn


@pytest.fixture
def sharepoint_list():
    """Create a SharepointList instance with patched connection."""
    with patch.object(SharepointList, "get_connection") as mock_get_conn:
        sp_list = SharepointList(credentials=DUMMY_CREDS)
        yield sp_list, mock_get_conn


def test_sharepoint_list_to_df_positive(
    sharepoint_list, mock_connection, mock_response
):
    """Test successful conversion of SharePoint list data to DataFrame."""
    sp_list, mock_get_conn = sharepoint_list

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.return_value = mock_response

    # Configure the mock response with a small dataset
    mock_response.json.return_value = {
        "d": {
            "results": [
                {"ID": 1, "Title": "Item 1"},
                {"ID": 2, "Title": "Item 2"},
                {"ID": 3, "Title": "Item 3"},
            ]
        }
    }

    # Call the function
    df = sp_list.to_df(
        list_site=LIST_SITE,
        list_name=LIST_NAME,
        select=["ID", "Title"],
        query=None,
    )

    # Basic assertions
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 3
    assert set(df.columns) >= {"ID", "Title"}

    # Verify the mock was called correctly
    expected_url = f"https://test.sharepoint.com/sites/{LIST_SITE}/_api/web/lists/GetByTitle('{LIST_NAME}')/items"
    mock_connection.get.assert_called_once()
    assert mock_connection.get.call_args[0][0] == expected_url


def test_sharepoint_list_to_df_no_items(
    sharepoint_list, mock_connection, mock_response
):
    """Test case when SharePoint list returns empty results."""
    sp_list, mock_get_conn = sharepoint_list

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.return_value = mock_response

    # Configure empty response
    mock_response.json.return_value = {"d": {"results": []}}

    # Test that appropriate exception is raised
    with pytest.raises(
        ValueError, match=f"No items found in SharePoint list {LIST_NAME}"
    ):
        sp_list.to_df(
            list_site=LIST_SITE,
            list_name=LIST_NAME,
            select=["ID", "Title"],
            query=None,
        )


def test_sharepoint_list_authentication_failure():
    """Test handling of authentication failure."""
    # Use context manager to ensure proper cleanup
    with patch("sharepy.connect") as mock_connect:
        mock_connect.side_effect = AuthError("Authentication failed")

        sp_list = SharepointList(credentials=DUMMY_CREDS)

        with pytest.raises(
            CredentialError,
            match="Could not authenticate to tenant.sharepoint.com with provided credentials.",
        ):
            sp_list.get_connection()


def test_sharepoint_list_missing_credentials():
    """Test validation of missing credentials."""
    incomplete_credentials = {**DUMMY_CREDS, "site": ""}

    with pytest.raises(
        CredentialError,
        match="'site', 'username', and 'password' credentials are required.",
    ):
        SharepointCredentials(**incomplete_credentials)


def test_sharepoint_list_to_df_query_filter(
    sharepoint_list, mock_connection, mock_response
):
    """Test applying a query filter when retrieving SharePoint list data."""
    sp_list, mock_get_conn = sharepoint_list

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.return_value = mock_response

    # Configure the mock response
    mock_response.json.return_value = {
        "d": {
            "results": [
                {"ID": 1, "Title": "Item 1"},
            ]
        }
    }

    # Call the function with a query
    df = sp_list.to_df(
        list_site=LIST_SITE,
        list_name=LIST_NAME,
        select=["ID", "Title"],
        query="Title eq 'Item 1'",
    )

    # Verify results
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 1
    assert df["Title"].iloc[0] == "Item 1"

    # Verify proper URL construction with query parameter
    expected_url = f"https://test.sharepoint.com/sites/{LIST_SITE}/_api/web/lists/GetByTitle('{LIST_NAME}')/items"
    mock_connection.get.assert_called_once()
    assert mock_connection.get.call_args[0][0] == expected_url

    # Verify the query parameter is correctly passed in the request parameters
    params = mock_connection.get.call_args[1]["params"]
    assert "$filter" in params
    assert params["$filter"] == "Title eq 'Item 1'"
    assert "$select" in params
    assert params["$select"] == "ID,Title"


def test_rename_case_insensitive_duplicated_columns():
    """Test handling of case-insensitive duplicate column names."""
    # Create a small test DataFrame
    df = pd.DataFrame({"ID": [1, 2], "id": [3, 4], "Title": ["A", "B"]})

    # Create minimal SharepointList instance
    with patch.object(SharepointList, "get_connection"):
        sp_list = SharepointList(credentials=DUMMY_CREDS)

        # Test the method
        rename_dict = sp_list._find_and_rename_case_insensitive_duplicated_column_names(
            df
        )

        # Verify results
        assert rename_dict
        assert len(rename_dict) == 2
        assert "id" in rename_dict
        assert "ID" in rename_dict

        # Apply rename without creating another full DataFrame copy
        renamed_columns = {col: rename_dict.get(col, col) for col in df.columns}
        assert "id_1" in renamed_columns.values()
        assert "id_2" in renamed_columns.values()


def test_sharepoint_list_pagination(sharepoint_list, mock_connection):
    """Test pagination for large lists (>5000 items)."""
    sp_list, mock_get_conn = sharepoint_list

    # Create separate response objects for each pagination call
    mock_response_1 = MagicMock()
    mock_response_2 = MagicMock()
    mock_response_3 = MagicMock()

    # Create a separate connection mock
    mock_connection = MagicMock()
    mock_connection.site = "test.sharepoint.com"
    mock_connection.get = MagicMock(
        side_effect=[mock_response_1, mock_response_2, mock_response_3]
    )

    # Return our custom mock_connection
    mock_get_conn.return_value = mock_connection

    # Configure mock responses with paginated data
    mock_response_1.json.return_value = {
        "d": {
            "results": [{"ID": i, "Title": f"Item {i}"} for i in range(1, 5001)],
            "__next": "next_page_url",
        }
    }
    mock_response_2.json.return_value = {
        "d": {
            "results": [{"ID": i, "Title": f"Item {i}"} for i in range(5001, 10001)],
            "__next": "next_page_url",
        }
    }
    mock_response_3.json.return_value = {
        "d": {"results": [{"ID": i, "Title": f"Item {i}"} for i in range(10001, 15001)]}
    }

    # Call the function
    df = sp_list.to_df(
        list_site=LIST_SITE,
        list_name=LIST_NAME,
        select=["ID", "Title"],
        query=None,
    )

    # Verify correct number of API calls were made
    assert mock_connection.get.call_count == 3

    # Basic assertions
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 15000
    assert df["ID"].min() == 1
    assert df["ID"].max() == 15000


def test_sharepoint_list_complex_data_types(
    sharepoint_list, mock_connection, mock_response
):
    """Test handling of complex data types in SharePoint lists."""
    sp_list, mock_get_conn = sharepoint_list

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.return_value = mock_response

    # Configure the mock response with complex data types
    mock_response.json.return_value = {
        "d": {
            "results": [
                {"ID": 1, "Title": "Item 1", "ComplexField": {"SubField": "Value"}},
                {"ID": 2, "Title": "Item 2", "ComplexField": {"SubField": "Value"}},
            ]
        }
    }

    # Call the function
    df = sp_list.to_df(
        list_site=LIST_SITE,
        list_name=LIST_NAME,
        select=["ID", "Title", "ComplexField"],
        query=None,
    )

    # Basic assertions
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 2
    assert set(df.columns) >= {"ID", "Title", "ComplexField"}


def test_sharepoint_list_url_construction_different_protocols(
    sharepoint_list, mock_connection, mock_response
):
    """Test URL construction with different protocols (http vs https)."""
    sp_list, mock_get_conn = sharepoint_list

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.return_value = mock_response

    # Configure the mock response
    mock_response.json.return_value = {"d": {"results": [{"ID": 1, "Title": "Item 1"}]}}

    # Call the function with http protocol
    sp_list.default_protocol = "http://"
    sp_list.to_df(
        list_site=LIST_SITE,
        list_name=LIST_NAME,
        select=["ID", "Title"],
        query=None,
    )

    # Verify the mock was called correctly
    expected_url = f"http://test.sharepoint.com/sites/{LIST_SITE}/_api/web/lists/GetByTitle('{LIST_NAME}')/items"
    mock_connection.get.assert_called_once()
    assert mock_connection.get.call_args[0][0] == expected_url


def test_sharepoint_list_http_error_responses(sharepoint_list, mock_connection):
    """Test HTTP error responses with proper status code extraction."""
    sp_list, mock_get_conn = sharepoint_list

    # Create a mock response with a status code
    mock_response = MagicMock()
    mock_response.status_code = 404

    # Create an HTTP error with the response attached
    http_error = requests.exceptions.HTTPError("404 Client Error: Not Found")
    http_error.response = mock_response

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.side_effect = http_error

    # Test that the appropriate exception is raised with status code information
    with pytest.raises(
        ValueError,
        match=f"HTTP error 404 when retrieving data from SharePoint list '{LIST_NAME}'",
    ):
        sp_list.to_df(
            list_site=LIST_SITE,
            list_name=LIST_NAME,
            select=["ID", "Title"],
            query=None,
        )


def test_sharepoint_list_http_error_unknown_status(sharepoint_list, mock_connection):
    """Test HTTP error responses when status code is unavailable."""
    sp_list, mock_get_conn = sharepoint_list

    # Create an HTTP error with no response
    http_error = requests.exceptions.HTTPError("Generic HTTP Error")
    # No response attribute set

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.side_effect = http_error

    # Test that appropriate exception is raised with "unknown" status
    with pytest.raises(
        ValueError,
        match=f"HTTP error unknown when retrieving data from SharePoint list '{LIST_NAME}'",
    ):
        sp_list.to_df(
            list_site=LIST_SITE,
            list_name=LIST_NAME,
            select=["ID", "Title"],
            query=None,
        )


def test_sharepoint_list_timeout_handling(sharepoint_list, mock_connection):
    """Test timeout handling."""
    sp_list, mock_get_conn = sharepoint_list

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.side_effect = TimeoutError("Request timed out")

    # Test that appropriate exception is raised with more specific message
    with pytest.raises(
        ValueError,
        match=f"Request to SharePoint list '{LIST_NAME}' timed out: Request timed out",
    ):
        sp_list.to_df(
            list_site=LIST_SITE,
            list_name=LIST_NAME,
            select=["ID", "Title"],
            query=None,
        )


def test_build_sharepoint_endpoint(sharepoint_list):
    """Test the _build_sharepoint_endpoint helper method."""
    sp_list, _ = sharepoint_list

    site_url = "https://test.sharepoint.com"
    endpoint = sp_list._build_sharepoint_endpoint(site_url, LIST_SITE, LIST_NAME)

    expected = f"https://test.sharepoint.com/sites/{LIST_SITE}/_api/web/lists/GetByTitle('{LIST_NAME}')/items"
    assert endpoint == expected


def test_ensure_protocol(sharepoint_list):
    """Test the _ensure_protocol helper method."""
    sp_list, _ = sharepoint_list

    # Default protocol should be "https://"
    site_without_protocol = "test.sharepoint.com"
    site_with_protocol = sp_list._ensure_protocol(site_without_protocol)
    assert site_with_protocol == "https://test.sharepoint.com"

    # Should not modify URLs that already have a protocol
    site_already_with_protocol = "https://test.sharepoint.com"
    result = sp_list._ensure_protocol(site_already_with_protocol)
    assert result == site_already_with_protocol

    # Should handle different protocols
    sp_list.default_protocol = "http://"
    result = sp_list._ensure_protocol(site_without_protocol)
    assert result == "http://test.sharepoint.com"


def test_get_records(sharepoint_list, mock_connection, mock_response):
    """Test the _get_records helper method."""
    sp_list, mock_get_conn = sharepoint_list

    # Setup mocks
    mock_get_conn.return_value = mock_connection
    mock_connection.get.return_value = mock_response

    mock_response.json.return_value = {
        "d": {"results": [{"ID": 1, "Title": "Item 1"}], "__next": "next_page_url"}
    }

    url = "https://test.sharepoint.com/sites/my_site/_api/web/lists/GetByTitle('my_list')/items"
    params = {"$select": "ID,Title"}

    # Call the helper method
    items, next_url = sp_list._get_records(url, params)

    # Verify results
    assert len(items) == 1
    assert items[0]["ID"] == 1
    assert items[0]["Title"] == "Item 1"
    assert next_url == "next_page_url"

    # Verify the mock was called correctly
    mock_connection.get.assert_called_once_with(url, params=params)


def test_get_records_with_next_url_object(sharepoint_list, mock_connection):
    """Test the _get_records helper with next link as an object."""
    sp_list, mock_get_conn = sharepoint_list

    # Setup mocks
    mock_response = MagicMock()
    mock_get_conn.return_value = mock_connection
    mock_connection.get.return_value = mock_response

    # Next link as an object with URI
    mock_response.json.return_value = {
        "d": {
            "results": [{"ID": 1, "Title": "Item 1"}],
            "__next": {"uri": "next_page_url_object"},
        }
    }

    url = "https://test.sharepoint.com/sites/my_site/_api/web/lists/GetByTitle('my_list')/items"

    # Call the helper method
    items, next_url = sp_list._get_records(url)

    # Verify results
    assert len(items) == 1
    assert next_url == "next_page_url_object"


def test_paginate_list_data(sharepoint_list):
    """Test the _paginate_list_data helper method."""
    sp_list, _ = sharepoint_list

    # We'll mock __get_records directly to test pagination logic
    with patch.object(SharepointList, "_get_records") as mock_fetch:
        # Set up the mock to return different results for each call
        mock_fetch.side_effect = [
            # First call returns 2 items and a next link
            ([{"ID": 1}, {"ID": 2}], "next_page_url"),
            # Second call returns 3 items and no next link
            ([{"ID": 3}, {"ID": 4}, {"ID": 5}], None),
        ]
        # Call the paginate method
        results = sp_list._paginate_list_data(
            "initial_url", {"param": "value"}, LIST_NAME
        )

        # Verify results
        assert len(results) == 5
        assert [item["ID"] for item in results] == [1, 2, 3, 4, 5]

        # Verify mock calls
        assert mock_fetch.call_count == 2
        mock_fetch.assert_any_call("initial_url", {"param": "value"})
        mock_fetch.assert_any_call("next_page_url", None)



import unittest

import pytest

from viadot.sources.vid_club import ValidationError, VidClub


class TestVidClub(unittest.TestCase):
    def setUp(self):
        """Setup VidClub instance before each test."""
        # Sample input data for the constructor
        self.vid_club = VidClub(
            endpoint="jobs", vid_club_credentials={"token": "test-token"}
        )

    def test_build_query(self):
        """Test correct URL generation for the 'jobs' endpoint."""
        # Sample input data for the build_query method
        from_date = "2023-01-01"
        to_date = "2023-01-31"
        api_url = "https://example.com/api/"
        items_per_page = 50
        endpoint = "jobs"
        region = "pl"

        # Expected result URL
        expected_url = "https://example.com/api/jobs?from=2023-01-01&to=2023-01-31&region=pl&limit=50"

        # Check if the method returns the correct URL
        result_url = self.vid_club.build_query(
            from_date, to_date, api_url, items_per_page, endpoint, region
        )
        assert result_url == expected_url

    def test_intervals(self):
        """Test breaking date range into intervals based on the days_interval."""
        # Sample input data for the intervals method
        from_date = "2023-01-01"
        to_date = "2023-01-15"
        days_interval = 5

        # Expected starts and ends lists
        expected_starts = ["2023-01-01", "2023-01-06", "2023-01-11"]
        expected_ends = ["2023-01-06", "2023-01-11", "2023-01-15"]

        # Check if the method returns correct intervals
        starts, ends = self.vid_club.intervals(from_date, to_date, days_interval)
        assert starts == expected_starts
        assert ends == expected_ends

    def test_intervals_invalid_date_range(self):
        """Test that ValidationError is raised when to_date is before from_date."""
        # Sample input data where to_date is before from_date
        from_date = "2023-01-15"
        to_date = "2023-01-01"
        days_interval = 5

        # Check if ValidationError is raised
        with pytest.raises(ValidationError):
            self.vid_club.intervals(from_date, to_date, days_interval)



"""Tasks for interacting with the Exchange Rates API."""

from datetime import datetime
from typing import Literal

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import ExchangeRates


Currency = Literal[
    "USD", "EUR", "GBP", "CHF", "PLN", "DKK", "COP", "CZK", "SEK", "NOK", "ISK"
]


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def exchange_rates_to_df(
    currency: Currency = "USD",
    credentials_secret: str | None = None,
    config_key: str | None = None,
    start_date: str = datetime.today().strftime("%Y-%m-%d"),
    end_date: str = datetime.today().strftime("%Y-%m-%d"),
    symbols: list[str] | None = None,
    tests: dict | None = None,
) -> pd.DataFrame:
    """Loads exchange rates from the Exchange Rates API into a pandas DataFrame.

    Args:
        currency (Currency, optional): Base currency to which prices of searched
            currencies are related. Defaults to "USD".
        credentials_secret (str, optional): The name of the secret storing
            the credentials. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        config_key (str, optional): The key in the viadot config holding relevant
            credentials.
            Defaults to None.
        start_date (str, optional): Initial date for data search.
            Data range is start_date -> end_date,
            supported format 'yyyy-mm-dd'.
            Defaults to datetime.today().strftime("%Y-%m-%d").
        end_date (str, optional): See above.
            Defaults to datetime.today().strftime("%Y-%m-%d").
        symbols (list[str], optional): List of ISO codes of currencies for which
            exchange rates from base currency will be fetched. Defaults to
            ["USD","EUR","GBP","CHF","PLN","DKK","COP","CZK","SEK","NOK","ISK"].
        tests (dict[str], optional): A dictionary with optional list of tests
            to verify the output dataframe. If defined, triggers the `validate` function
            from viadot.utils. Defaults to None.

    Returns:
        pd.DataFrame: The pandas `DataFrame` containing data from the file.
    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    if not symbols:
        symbols = [
            "USD",
            "EUR",
            "GBP",
            "CHF",
            "PLN",
            "DKK",
            "COP",
            "CZK",
            "SEK",
            "NOK",
            "ISK",
        ]

    if not config_key:
        credentials = get_credentials(credentials_secret)

    e = ExchangeRates(
        currency=currency,
        start_date=start_date,
        end_date=end_date,
        symbols=symbols,
        credentials=credentials,
        config_key=config_key,
    )
    return e.to_df(tests=tests)



"""Task for downloading data from TM1 to a pandas DataFrame."""

from pandas import DataFrame
from prefect import task
from prefect.logging import get_run_logger

from viadot.config import get_source_credentials
from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources.tm1 import TM1


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60 * 3)
def tm1_to_df(
    mdx_query: str | None = None,
    cube: str | None = None,
    view: str | None = None,
    limit: int | None = None,
    private: bool = False,
    credentials_secret: str | None = None,
    config_key: str | None = None,
    verify: bool = False,
    if_empty: str = "skip",
) -> DataFrame:
    """Download data from TM1 to pandas DataFrame.

    Args:
        mdx_query (str, optional): MDX select query needed to download the data.
            Defaults to None.
        cube (str, optional): Cube name from which data will be downloaded.
            Defaults to None.
        view (str, optional): View name from which data will be downloaded.
            Defaults to None.
        limit (str, optional): How many rows should be extracted.
            If None all the available rows will be downloaded. Defaults to None.
        private (bool, optional): Whether or not data download should be private.
            Defaults to False.
        credentials_secret (dict[str, Any], optional): The name of the secret that
            stores TM1 credentials.
            More info on: https://docs.prefect.io/concepts/blocks/. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to "TM1".
        verify (bool, optional): Whether or not verify SSL certificate.
            Defaults to False.
        if_empty (str, optional): What to do if output DataFrame is empty.
            Defaults to "skip".

    """
    if not (credentials_secret or config_key):
        raise MissingSourceCredentialsError

    logger = get_run_logger()

    credentials = get_source_credentials(config_key) or get_credentials(
        credentials_secret
    )

    bc = TM1(
        credentials=credentials,
        config_key=config_key,
        limit=limit,
        private=private,
        verify=verify,
    )

    df = bc.to_df(
        if_empty=if_empty,
        mdx_query=mdx_query,
        cube=cube,
        view=view,
    )

    nrows = df.shape[0]
    ncols = df.shape[1]

    logger.info(
        f"Successfully downloaded {nrows} rows and {ncols} columns of data to a DataFrame."
    )

    return df



"""Flows for downloading data from SAP to Parquet file."""

from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import sap_rfc_to_df
from viadot.orchestration.prefect.tasks.task_utils import df_to_parquet


@flow(
    name="extract--sap--parquet",
    description="Extract data from SAP and load it into Parquet file",
    retries=1,
    retry_delay_seconds=60,
)
def sap_to_parquet(
    path: str,
    if_exists: Literal["append", "replace", "skip"] = "replace",
    query: str | None = None,
    func: str | None = None,
    sap_sep: str | None = None,
    rfc_total_col_width_character_limit: int = 400,
    rfc_unique_id: list[str] | None = None,
    sap_credentials_secret: str | None = None,
    sap_config_key: str = "SAP",
    replacement: str = "-",
) -> None:
    """Download a pandas `DataFrame` from SAP load it into Parquet file.

    Args:
        path (str): Path to Parquet file, where the data will be located.
            Defaults to None.
        if_exists (Literal["append", "replace", "skip"], optional): What to do if the
            file exists. Defaults to "replace".
        query (str): The query to be executed with pyRFC.
        sap_sep (str, optional): The separator to use when reading query results.
            If not provided, multiple options are automatically tested.
            Defaults to None.
        func (str, optional): SAP RFC function to use. Defaults to None.
        rfc_total_col_width_character_limit (int, optional): Number of characters by
            which query will be split in chunks in case of too many columns for RFC
            function. According to SAP documentation, the limit is 512 characters.
            However, it was observed that SAP raising an exception even on a slightly
            lower number of characters, so safety margin was added. Defaults to 400.
        rfc_unique_id  (list[str], optional): Reference columns to merge chunks Data
            Frames. These columns must to be unique. Otherwise, the table will be
            malformed. If no columns are provided, all data frame columns will be
            concatenated. Defaults to None.
        sap_credentials_secret (str, optional): The name of the Prefect Secret that
            stores SAP credentials. Defaults to None.
        sap_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to "SAP".
        replacement (str, optional): In case of sep is on a columns, set up a new
            character to replace inside the string to avoid flow breakdowns.
            Defaults to "-".
    """
    df = sap_rfc_to_df(
        query=query,
        sep=sap_sep,
        func=func,
        replacement=replacement,
        rfc_total_col_width_character_limit=rfc_total_col_width_character_limit,
        rfc_unique_id=rfc_unique_id,
        config_key=sap_config_key,
        credentials_secret=sap_credentials_secret,
    )

    return df_to_parquet(
        df=df,
        path=path,
        if_exists=if_exists,
    )



"""Source for connecting to Supermetrics API."""

import json
from typing import Any

import numpy as np
import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns, handle_api_response


class SupermetricsCredentials(BaseModel):
    """Represents credentials for accessing the Supermetrics API.

    This class encapsulates the necessary credentials required to authenticate
    and access the Supermetrics API.

    Attributes:
    ----------
        user (str):
            The email account associated with the Supermetrics user.
        api_key (str):
            The API key that provides access to the Supermetrics API.

    """

    user: str
    api_key: str


class Supermetrics(Source):
    """Implements methods for querying and interacting with the Supermetrics API.

    This class provides functionality to query data from the Supermetrics API,
    which is a tool used for accessing data from various data sources. The API
    documentation can be found at:
    https://supermetrics.com/docs/product-api-getting-started/. For information
    on usage limits, please refer to:
    https://supermetrics.com/docs/product-api-usage-limits/.

    Args:
    ----
        config_key (str, optional):
            The key in the viadot configuration that holds the relevant credentials
            for the API. Defaults to None.
        credentials (dict of str to any, optional):
            A dictionary containing the credentials needed for the API connection
            configuration, specifically `api_key` and `user`. Defaults to None.
        query_params (dict of str to any, optional):
            A dictionary containing the parameters to pass to the GET query.
            These parameters define the specifics of the data request. Defaults to None.
            For a full specification of possible parameters, see:
            https://supermetrics.com/docs/product-api-get-data/.

    """

    BASE_URL = "https://api.supermetrics.com/enterprise/v2/query/data/json"

    def __init__(
        self,
        *args,
        credentials: dict[str, Any] | None = None,
        config_key: str | None = None,
        query_params: dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        """Initialize the Supermetrics object.

        This constructor sets up the necessary components to interact with the
        Supermetrics API, including the credentials and any query parameters.

        Args:
        ----
            credentials (SupermetricsCredentials, optional):
                An instance of `SupermetricsCredentials` containing the API key and
                user email for authentication. Defaults to None.
            config_key (str, optional):
                The key in the viadot configuration that holds the relevant credentials
                for the API. Defaults to None.
            query_params (dict of str to any, optional):
                A dictionary containing the parameters to pass to the GET query. These
                parameters define the specifics of the data request. Defaults to None.

        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(SupermetricsCredentials(**raw_creds))

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.query_params = query_params

    def to_json(self, timeout: tuple = (3.05, 60 * 30)) -> dict[str, Any]:
        """Download query results to a dictionary.

        This method executes the query against the Supermetrics API and retrieves
        the results as a JSON dictionary.

        Args:
        ----
            timeout (tuple of float, optional):
                A tuple specifying the timeout values for the request. The first value
                is the timeout for connection issues, and the second value is
                the timeout for query execution. Defaults to (3.05, 1800), which
                provides a short timeout for connection issues and a longer timeout
                for the query execution.

        Returns:
        -------
            dict:
                The response from the Supermetrics API, returned as a JSON dictionary.

        Raises:
        ------
            ValueError:
                Raised if the query parameters are not set before calling this method.

        """
        if not self.query_params:
            msg = "Please build the query first"
            raise ValueError(msg)

        params = {"json": json.dumps(self.query_params)}
        headers = {"Authorization": f"Bearer {self.credentials.get('api_key')}"}

        response = handle_api_response(
            url=Supermetrics.BASE_URL,
            params=params,
            headers=headers,
            timeout=timeout,
        )
        return response.json()

    @classmethod
    def _get_col_names_google_analytics(
        cls,
        response: dict,
    ) -> list[str]:
        """Get column names from Google Analytics data.

        This method extracts the column names from the JSON response received
        from a Google Analytics API call.

        Args:
        ----
            response (dict):
                A dictionary containing the JSON response from the API call.

        Returns:
        -------
            list of str:
                A list of column names extracted from the Google Analytics data.

        Raises:
        ------
            ValueError:
                Raised if no data is returned in the response or if the column names
                cannot be determined.

        """
        is_pivoted = any(
            field["field_split"] == "column"
            for field in response["meta"]["query"]["fields"]
        )

        if is_pivoted:
            if not response["data"]:
                msg = "Couldn't find column names as query returned no data."
                raise ValueError(msg)
            columns = response["data"][0]
        else:
            cols_meta = response["meta"]["query"]["fields"]
            columns = [col_meta["field_name"] for col_meta in cols_meta]
        return columns

    @classmethod
    def _get_col_names_other(cls, response: dict) -> list[str]:
        """Get column names from non-Google Analytics data.

        This method extracts the column names from the JSON response received
        from an API call that is not related to Google Analytics.

        Args:
        ----
            response (dict):
                A dictionary containing the JSON response from the API call.

        Returns:
        -------
            list of str:
                A list of column names extracted from the non-Google Analytics data.

        """
        cols_meta = response["meta"]["query"]["fields"]
        return [col_meta["field_name"] for col_meta in cols_meta]

    def _get_col_names(self) -> list[str]:
        """Get column names based on the data type.

        This method determines the appropriate column names for the data based
        on its type, whether it's Google Analytics data or another type.

        Returns:
        -------
            list of str:
                A list of column names based on the data type.

        Raises:
        ------
            ValueError:
                Raised if the column names cannot be determined.

        """
        response: dict = self.to_json()
        if self.query_params["ds_id"] == "GA":
            return Supermetrics._get_col_names_google_analytics(response)

        return Supermetrics._get_col_names_other(response)

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: str = "warn",
        query_params: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Download data into a pandas DataFrame.

        This method retrieves data from the Supermetrics API and loads it into
        a pandas DataFrame.

        Args:
        ----
            if_empty (str, optional):
                Specifies the action to take if the query returns no data.
                Options include "fail" to raise an error or "ignore" to return
                an empty DataFrame. Defaults to "fail".

        Returns:
        -------
            pd.DataFrame:
                A pandas DataFrame containing the JSON data retrieved from the API.

        Raises:
        ------
            ValueError:
                Raised if the DataFrame is empty and `if_empty` is set to "fail".

        """
        # Use provided query_params or default to the instance's query_params
        if query_params is not None:
            self.query_params = query_params

        if not self.query_params:
            msg = "Query parameters are required to fetch data."
            raise ValueError(msg)

        self.query_params["api_key"] = self.credentials.get("api_key")

        try:
            columns = self._get_col_names()
        except ValueError:
            columns = None

        data = self.to_json()["data"]

        if data:
            df = pd.DataFrame(data[1:], columns=columns).replace("", np.nan)
        else:
            df = pd.DataFrame(columns=columns)

        if df.empty:
            self._handle_if_empty(if_empty)

        return df



import pandas as pd
import pytest

from viadot.config import get_source_credentials
from viadot.exceptions import DataRangeError
from viadot.sources import Epicor


FILTERS_NOK = """
    <OrderQuery>
        <QueryFields>
            <CompanyNumber>001</CompanyNumber>
            <BegInvoiceDate></BegInvoiceDate>
            <EndInvoiceDate>2022-05-16</EndInvoiceDate>
            <RecordCount>3</RecordCount>
        </QueryFields>
    </OrderQuery>"""

FILTERS_ORDERS = """
    <OrderQuery>
        <QueryFields>
            <CompanyNumber>001</CompanyNumber>
            <BegInvoiceDate>2022-05-16</BegInvoiceDate>
            <EndInvoiceDate>2022-05-16</EndInvoiceDate>
            <RecordCount>3</RecordCount>
        </QueryFields>
    </OrderQuery>"""
FILTERS_BOOKINGS = """
    <BookingsQuery>
        <QueryFields>
            <CompanyNumber></CompanyNumber>
            <CustomerNumber></CustomerNumber>
            <SellingWarehouse></SellingWarehouse>
            <WrittenBy></WrittenBy>
            <SalesRepOne></SalesRepOne>
            <BegUpdateDate>2022-05-16</BegUpdateDate>
            <EndUpdateDate>2022-05-16</EndUpdateDate>
            <SortXMLTagName></SortXMLTagName>
            <SortMethod></SortMethod>
            <RecordCount></RecordCount>
            <RecordCountPage></RecordCountPage>
        </QueryFields>
    </BookingsQuery>"""
FILTERS_OPEN_ORDERS = """
    <OrderQuery>
        <QueryFields>
            <CompanyNumber></CompanyNumber>
            <CustomerNumber></CustomerNumber>
            <SellingWarehouse></SellingWarehouse>
            <OrderNumber></OrderNumber>
            <WrittenBy></WrittenBy>
            <CustomerPurchaseOrderNumber></CustomerPurchaseOrderNumber>
            <CustomerReleaseNumber></CustomerReleaseNumber>
            <CustomerJobNumber></CustomerJobNumber>
            <EcommerceId></EcommerceId>
            <EcommerceOrderNumber></EcommerceOrderNumber>
            <BegOrderDate></BegOrderDate>
            <EndOrderDate></EndOrderDate>
            <SortXMLTagName></SortXMLTagName>
            <SortMethod></SortMethod>
            <RecordCount></RecordCount>
            <RecordCountPage></RecordCountPage>
        </QueryFields>
    </OrderQuery>"""
FILTERS_CUSTOMERS = """
    <Query>
        <QueryFields>
            <CompanyNumber></CompanyNumber>
            <CustomerNumber></CustomerNumber>
            <SellingWarehouse></SellingWarehouse>
            <City></City>
            <State></State>
            <ZipCode></ZipCode>
            <EmailAddress></EmailAddress>
            <Phone></Phone>
            <BegLastChangeDate></BegLastChangeDate>
            <EndLastChangeDate></EndLastChangeDate>
            <SortXMLTagName>1</SortXMLTagName>
            <SortMethod></SortMethod>
            <RecordCount></RecordCount>
            <RecordCountPage></RecordCountPage>
        </QueryFields>
    </Query>"""


@pytest.fixture(scope="session")
def epicor():
    return Epicor(
        base_url=get_source_credentials("epicor").get("test_url"),
        config_key="epicor",
        validate_date_filter=False,
    )


def test_connection(epicor: Epicor):
    assert epicor.get_xml_response(FILTERS_ORDERS).ok


def test_validate_filter(epicor: Epicor):
    with pytest.raises(DataRangeError):
        epicor.validate_filter(FILTERS_NOK)


def test_to_df_orders(epicor: Epicor):
    df = epicor.to_df(FILTERS_ORDERS)
    assert isinstance(df, pd.DataFrame)


def test_to_df_bookings(epicor: Epicor):
    df = epicor.to_df(FILTERS_BOOKINGS)
    assert isinstance(df, pd.DataFrame)


def test_to_df_open_orders(epicor: Epicor):
    df = epicor.to_df(FILTERS_OPEN_ORDERS)
    assert isinstance(df, pd.DataFrame)


def test_to_df_customers(epicor: Epicor):
    df = epicor.to_df(FILTERS_CUSTOMERS)
    assert isinstance(df, pd.DataFrame)



"""SAP BW  connector."""

import textwrap
from typing import Any

import pandas as pd
from pydantic import BaseModel
import pyrfc

from viadot.config import get_source_credentials
from viadot.exceptions import ValidationError
from viadot.sources.base import Source
from viadot.utils import add_viadot_metadata_columns


class SAPBWCredentials(BaseModel):
    """Checking for values in SAP BW credentials dictionary.

    Two key values are held in the SAP BW connector:
        - ashost: Indicates the host name or IP address of a specific SAP
            application server.
        - client: Specifies the SAP logon parameter client.
        - passwd: Indicates the SAP logon parameter password.
        - sysnr: Indicates the SAP system number—the 2-byte code that identifies the
            system on the host.
        - user: Indicates the SAP logon parameter user.

    Args:
        BaseModel (pydantic.main.ModelMetaclass): A base class for creating
            Pydantic models.
    """

    ashost: str
    client: str
    passwd: str
    sysnr: str
    user: str


class SAPBW(Source):
    """Quering the SAP BW (SAP Business Warehouse) source using pyrfc library.

    Documentation to pyrfc can be found under:
        https://sap.github.io/PyRFC/pyrfc.html
    Documentation for SAP connection modules under:
        https://www.se80.co.uk/sap-function-modules/list/?index=rsr_mdx
    """

    def __init__(
        self,
        *args,
        credentials: SAPBWCredentials | None = None,
        config_key: str = "sap_bw",
        **kwargs,
    ):
        """Create an instance of SAP BW.

        Args:
            credentials (Optional[SAPBWCredentials], optional): SAP BW credentials.
                Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to "sap_bw".

        Examples:
            sap_bw = SAPBW(
                credentials=credentials,
                config_key=config_key,
            )
            sap_bw.api_connection(
                ...
            )
            data_frame = sap_bw.to_df()

        Raises:
            CredentialError: If credentials are not provided in local_config or
                directly as a parameter.
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(SAPBWCredentials(**raw_creds))

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.query_output = None

    def _create_connection(self):
        """Create the connection with SAP BW.

        Returns:
            Connection: Connection to SAP.
        """
        return pyrfc.Connection(
            ashost=self.credentials.get("ashost"),
            sysnr=self.credentials.get("sysnr"),
            user=self.credentials.get("user"),
            passwd=self.credentials.get("passwd"),
            client=self.credentials.get("client"),
        )

    def api_connection(self, mdx_query: str) -> None:
        """Generate the SAP BW output dataset from MDX query.

        Args:
            mdx_query (str): The MDX query to be passed to connection.
        """
        conn = self._create_connection()

        query = textwrap.wrap(mdx_query, 75)
        properties = conn.call("RSR_MDX_CREATE_OBJECT", COMMAND_TEXT=query)

        datasetid = properties["DATASETID"]
        self.query_output = conn.call("RSR_MDX_GET_FLAT_DATA", DATASETID=datasetid)
        conn.close()

    def _apply_user_mapping(
        self,
        df: pd.DataFrame,
        mapping_dict: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Apply the column mapping defined by user for the output dataframe.

            DataFrame will be cut to selected columns - if any other columns need to be
        included in the output file, please add them to the mapping dictionary with
        original names.

        Args:
            df (pd.DataFrame): Input dataframe for the column mapping task.
            mapping_dict (dict[str, Any], optional): Dictionary with original and new
                column names. Defaults to None.

        Returns:
            pd.DataFrame: Output DataFrame with mapped columns.
        """
        self.logger.info("Applying user defined mapping for columns...")
        df = df[mapping_dict.keys()]
        df.columns = mapping_dict.values()

        self.logger.info("Successfully applied user mapping.")

        return df

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: str = "warn",
        mapping_dict: dict[str, Any] | None = None,
    ) -> pd.DataFrame:
        """Convert the SAP BW output JSON data into a dataframe.

        Args:
            if_empty (str, optional): What to do if a fetch produce no data.
                Defaults to "warn".
            mapping_dict (dict[str, Any], optional): Dictionary with original and new
                column names. Defaults to None.

        Raises:
            ValidationError: Prints the original SAP error message in case of issues
                with MDX execution.

        Returns:
            pd.Dataframe: The response data as a pandas DataFrame, enriched
                with viadot metadata columns.
        """
        raw_data = {}

        if self.query_output["RETURN"]["MESSAGE"] == "":
            results = self.query_output["DATA"]
            for cell in results:
                if cell["ROW"] not in raw_data:
                    raw_data[cell["ROW"]] = {}
                if "].[" not in cell["DATA"]:
                    raw_data[cell["ROW"]][cell["COLUMN"]] = cell["DATA"]
            rows = [raw_data[row] for row in raw_data]
            cols = [x["DATA"] for x in self.query_output["HEADER"]]

            data_frame = pd.DataFrame(data=rows)
            data_frame.columns = cols

        else:
            data_frame = pd.DataFrame()
            raise ValidationError(self.query_output["RETURN"]["MESSAGE"])

        if mapping_dict:
            data_frame = self._apply_user_mapping(data_frame, mapping_dict)

        if data_frame.empty:
            self._handle_if_empty(
                if_empty=if_empty,
                message="The response does not contain any data.",
            )
        else:
            self.logger.info("Successfully downloaded data from the Mindful API.")

        return data_frame

    def get_available_columns(self, mdx_query: str) -> list[str]:
        """Generate list of all available columns in a SAP BW table.

        Args:
            mdx_query (str): The MDX query to be passed to connection.

        Returns:
            list[str]: List of all available columns in the source table.
        """
        conn = self._create_connection()
        query = textwrap.wrap(mdx_query, width=75)

        properties = conn.call("RSR_MDX_CREATE_STORED_OBJECT", COMMAND_TEXT=query)
        datasetid = properties["DATASETID"]

        if properties["RETURN"]["MESSAGE"] == "":
            get_axis_info = conn.call("RSR_MDX_GET_AXIS_INFO", DATASETID=datasetid)
            cols = get_axis_info["AXIS_DIMENSIONS"]

            all_available_columns = [x["DIM_UNAM"] for x in cols]
        else:
            all_available_columns = []
            self.logger.error(properties["RETURN"]["MESSAGE"])

        return all_available_columns



from unittest.mock import patch

import pandas as pd
import pytest

from viadot.orchestration.prefect.flows import azure_sql_to_adls
from viadot.sources.azure_data_lake import AzureDataLake


@pytest.fixture
def query():
    return "SELECT * FROM your_table_name"


@pytest.fixture
def TEST_FILE_PATH():
    return "test_file_path"


@pytest.fixture
def adls_credentials_secret():
    return "mock_adls_credentials_secret"


@pytest.fixture
def azure_sql_credentials_secret():
    return "mock_azure_sql_credentials_secret"


def test_azure_sql_to_adls(
    query,
    TEST_FILE_PATH,
    adls_credentials_secret,
    azure_sql_credentials_secret,
):
    lake = AzureDataLake(config_key="adls_test")

    # Ensure the file does not exist before the test
    assert not lake.exists(TEST_FILE_PATH)

    with (
        patch(
            "viadot.orchestration.prefect.tasks.azure_sql_to_df"
        ) as mock_azure_sql_to_df,
        patch("viadot.orchestration.prefect.tasks.df_to_adls") as mock_df_to_adls,
    ):
        mock_df = pd.DataFrame({"column1": [1, 2], "column2": [3, 4]})
        mock_azure_sql_to_df.return_value = mock_df

        # Call the flow
        azure_sql_to_adls(
            query=query,
            credentials_secret=azure_sql_credentials_secret,
            validate_df_dict=None,
            convert_bytes=False,
            remove_special_characters=None,
            columns_to_clean=None,
            adls_config_key="adls_test",
            adls_azure_key_vault_secret=adls_credentials_secret,
            adls_path=TEST_FILE_PATH,
            adls_path_overwrite=True,
        )

        # Assert that the azure_sql_to_df task was called with the correct arguments
        mock_azure_sql_to_df.assert_called_once_with(
            query=query,
            credentials_secret=azure_sql_credentials_secret,
            sep=",",
            file_path=TEST_FILE_PATH,
            if_exists="replace",
            validate_df_dict=None,
            convert_bytes=False,
            remove_special_characters=None,
            columns_to_clean=None,
        )

        # Assert that df_to_adls was called with the correct arguments
        mock_df_to_adls.assert_called_once_with(
            df=mock_df,
            path=TEST_FILE_PATH,
            credentials_secret=adls_credentials_secret,
            config_key="adls_test",
            overwrite=True,
        )

    assert lake.exists(TEST_FILE_PATH)

    lake.rm(TEST_FILE_PATH)



"""'test_salesforce.py'."""

from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
from simple_salesforce import Salesforce as SimpleSalesforce

from viadot.exceptions import CredentialError
from viadot.sources import Salesforce
from viadot.sources.salesforce import SalesforceCredentials


variables = {
    "credentials": {
        "username": "test_user",
        "password": "test_password",  # pragma: allowlist secret
        "token": "test_token",
    },
    "records_1": [
        {
            "Id": "001",
            "Name": "Test Record",
            "attributes": {
                "type": "Account",
                "url": "/services/data/v50.0/sobjects/Account/001",
            },
        },
    ],
    "records_2": [
        {
            "Id": "001",
            "Name": "Test Record",
            "attributes": {
                "type": "Account",
                "url": "/services/data/v50.0/sobjects/Account/001",
            },
        },
    ],
    "data": [
        {
            "Id": "001",
            "Name": "Test Record",
            "attributes": {
                "type": "Account",
                "url": "/services/data/v50.0/sobjects/Account/001",
            },
        },
        {
            "Id": "002",
            "Name": "Another Record",
            "attributes": {
                "type": "Account",
                "url": "/services/data/v50.0/sobjects/Account/001",
            },
        },
    ],
}


@pytest.fixture
def mock_sf_instance(mocker):
    """Fixture to mock the SimpleSalesforce instance."""
    mock_sf_instance = mocker.MagicMock(spec=SimpleSalesforce)

    mocker.patch(
        "viadot.sources.salesforce.SimpleSalesforce", return_value=mock_sf_instance
    )
    return mock_sf_instance


@pytest.mark.basic
def test_salesforce_init_dev_env(mock_sf_instance):
    """Test Salesforce, starting in dev mode."""
    sf_instance = Salesforce(credentials=variables["credentials"], env="DEV")

    assert sf_instance.salesforce == mock_sf_instance


class TestSalesforceCredentials:
    """Test Salesforce Credentials Class."""

    @pytest.mark.basic
    def test_salesforce_credentials(self):
        """Test Salesforce credentials."""
        SalesforceCredentials(
            username="test_user",
            password="test_password",  # noqa: S106 # pragma: allowlist secret
            token="test_token",  # noqa: S106
        )


@pytest.mark.basic
def test_salesforce_init_prod_env(mock_sf_instance):
    """Test Salesforce, starting in prod mode."""
    sf_instance = Salesforce(credentials=variables["credentials"], env="PROD")

    assert sf_instance.salesforce == mock_sf_instance


@pytest.mark.basic
def test_salesforce_invalid_env():
    """Test Salesforce, invalid `env` parameter."""
    with pytest.raises(
        ValueError, match="The only available environments are DEV, QA, and PROD."
    ):
        Salesforce(credentials=variables["credentials"], env="INVALID")


@pytest.mark.basic
def test_salesforce_missing_credentials():
    """Test Salesforce missing credentials."""
    incomplete_creds = {
        "username": "user",  # pragma: allowlist secret
        "password": "pass",  # pragma: allowlist secret
    }
    with pytest.raises(CredentialError):
        Salesforce(credentials=incomplete_creds)


@pytest.mark.connect
def test_salesforce_to_df_with_columns(mock_sf_instance):
    """Test Salesforce `to_df` method with columns."""
    salesforce_instance = Salesforce(credentials=variables["credentials"])

    mock_sf_instance.query.return_value = {"records": variables["records_2"]}

    result_df = salesforce_instance.to_df(table="Account", columns=["Id", "Name"])

    result_df.drop(
        columns=["_viadot_source", "_viadot_downloaded_at_utc"],
        inplace=True,
        axis=1,
    )

    pd.testing.assert_frame_equal(
        result_df, pd.DataFrame([{"Id": "001", "Name": "Test Record"}])
    )
    mock_sf_instance.query.assert_called_once_with("SELECT Id, Name FROM Account")


@pytest.mark.functions
def test_salesforce_to_df(mock_sf_instance):
    """Test Salesforce `to_df` method."""
    salesforce_instance = Salesforce(credentials=variables["credentials"])
    mock_sf_instance.query.return_value = {"records": variables["data"]}

    df = salesforce_instance.to_df()
    assert not df.empty
    assert df.shape == (2, 4)
    assert list(df.columns) == [
        "Id",
        "Name",
        "_viadot_source",
        "_viadot_downloaded_at_utc",
    ]
    assert df.iloc[0]["Id"] == "001"


@pytest.mark.functions
def test_salesforce_to_df_empty_data(mock_sf_instance):  # noqa: ARG001
    """Test Salesforce `to_df` method with empty df."""
    salesforce_instance = Salesforce(credentials=variables["credentials"])
    salesforce_instance.data = []

    with pytest.raises(ValueError, match="The response does not contain any data."):
        salesforce_instance.to_df(if_empty="fail")


@pytest.mark.functions
def test_salesforce_to_df_warn_empty_data(mock_sf_instance):  # noqa: ARG001
    """Test Salesforce `to_df` method with empty df, warn."""
    salesforce_instance = Salesforce(credentials=variables["credentials"])
    salesforce_instance.data = []

    df = salesforce_instance.to_df(if_empty="warn")

    assert df.empty


@pytest.mark.functions
def test_salesforce_upsert_with_empty_data():
    with patch("viadot.sources.salesforce.SimpleSalesforce") as mock_salesforce_class:
        mock_salesforce_instance = mock_salesforce_class.return_value
        mock_salesforce_instance.upsert.return_value = None

        df = pd.DataFrame(columns=["a", "b", "c"])
        salesforce_instance = mock_salesforce_class(variables["credentials"])

        result = salesforce_instance.upsert(df=df, table="Contact", external_id="a")
        assert result is None


@pytest.mark.functions
def test_salesforce_upsert_with_missing_external_key(mock_sf_instance):  # noqa: ARG001
    with patch.object(SimpleSalesforce, "salesforce", create=True):
        sf = Salesforce(credentials=variables["credentials"])

        sf.salesforce = MagicMock()
        sf.salesforce.test = MagicMock()

        df = pd.DataFrame([{"a": 1, "b": "foo"}, {"a": 2, "b": "bar"}])

        with pytest.raises(
            ValueError, match="Passed DataFrame does not contain column 'c'."
        ):
            sf.upsert(df=df, table="test", external_id="c")



"""Import flows."""

from .azure_sql_to_adls import azure_sql_to_adls
from .bigquery_to_adls import bigquery_to_adls
from .business_core_to_parquet import business_core_to_parquet
from .cloud_for_customers_to_adls import cloud_for_customers_to_adls
from .cloud_for_customers_to_databricks import cloud_for_customers_to_databricks
from .customer_gauge_to_adls import customer_gauge_to_adls
from .duckdb_to_parquet import duckdb_to_parquet
from .duckdb_to_sql_server import duckdb_to_sql_server
from .duckdb_transform import duckdb_transform
from .epicor_to_parquet import epicor_to_parquet
from .eurostat_to_adls import eurostat_to_adls
from .exchange_rates_to_adls import exchange_rates_to_adls
from .exchange_rates_to_databricks import exchange_rates_to_databricks
from .exchange_rates_to_redshift_spectrum import exchange_rates_api_to_redshift_spectrum
from .genesys_to_adls import genesys_to_adls
from .hubspot_to_adls import hubspot_to_adls
from .mediatool_to_adls import mediatool_to_adls
from .mindful_to_adls import mindful_to_adls
from .outlook_to_adls import outlook_to_adls
from .salesforce_to_adls import salesforce_to_adls
from .sap_bw_to_adls import sap_bw_to_adls
from .sap_to_parquet import sap_to_parquet
from .sap_to_redshift_spectrum import sap_to_redshift_spectrum
from .sftp_to_adls import sftp_to_adls
from .sharepoint_list_to_redshift_spectrum import sharepoint_list_to_redshift_spectrum
from .sharepoint_to_adls import sharepoint_to_adls
from .sharepoint_to_databricks import sharepoint_to_databricks
from .sharepoint_to_redshift_spectrum import sharepoint_to_redshift_spectrum
from .sharepoint_to_s3 import sharepoint_to_s3
from .sql_server_to_minio import sql_server_to_minio
from .sql_server_to_parquet import sql_server_to_parquet
from .sql_server_transform import sql_server_transform
from .supermetrics_to_adls import supermetrics_to_adls
from .tm1_to_parquet import tm1_to_parquet
from .transform import transform
from .transform_and_catalog import transform_and_catalog
from .vid_club_to_adls import vid_club_to_adls


__all__ = [
    "azure_sql_to_adls",
    "bigquery_to_adls",
    "business_core_to_parquet",
    "cloud_for_customers_to_adls",
    "cloud_for_customers_to_databricks",
    "customer_gauge_to_adls",
    "duckdb_to_parquet",
    "duckdb_to_sql_server",
    "duckdb_transform",
    "epicor_to_parquet",
    "eurostat_to_adls",
    "exchange_rates_api_to_redshift_spectrum",
    "exchange_rates_to_adls",
    "exchange_rates_to_databricks",
    "genesys_to_adls",
    "hubspot_to_adls",
    "mediatool_to_adls",
    "mindful_to_adls",
    "outlook_to_adls",
    "salesforce_to_adls",
    "sap_bw_to_adls",
    "sap_to_parquet",
    "sap_to_redshift_spectrum",
    "sftp_to_adls",
    "sharepoint_list_to_redshift_spectrum",
    "sharepoint_to_adls",
    "sharepoint_to_databricks",
    "sharepoint_to_redshift_spectrum",
    "sharepoint_to_s3",
    "sql_server_to_minio",
    "sql_server_to_parquet",
    "sql_server_transform",
    "supermetrics_to_adls",
    "tm1_to_parquet",
    "transform",
    "transform_and_catalog",
    "vid_club_to_adls",
]



"""'bigquery_to_adls.py'."""

from prefect import flow
from prefect.task_runners import ConcurrentTaskRunner

from viadot.orchestration.prefect.tasks import bigquery_to_df, df_to_adls


@flow(
    name="BigQuery extraction to ADLS",
    description="Extract data from BigQuery and load it into Azure Data Lake Storage.",
    retries=1,
    retry_delay_seconds=60,
    task_runner=ConcurrentTaskRunner,
)
def bigquery_to_adls(  # noqa: PLR0913
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    query: str | None = None,
    dataset_name: str | None = None,
    table_name: str | None = None,
    date_column_name: str | None = None,
    start_date: str | None = None,
    end_date: str | None = None,
    columns: list[str] | None = None,
    adls_config_key: str | None = None,
    adls_azure_key_vault_secret: str | None = None,
    adls_path: str | None = None,
    adls_path_overwrite: bool = False,
) -> None:
    """Download data from BigQuery to Azure Data Lake.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (str, optional): The name of the Azure Key Vault secret
            where credentials are stored. Defaults to None.
        query (str, optional): SQL query to querying data in BigQuery. Format of basic
            query: (SELECT * FROM `{project}.{dataset_name}.{table_name}`).
            Defaults to None.
        dataset_name (str, optional): Dataset name. Defaults to None.
        table_name (str, optional): Table name. Defaults to None.
        dataset_name (str, optional): Dataset name. Defaults to None.
        date_column_name (str, optional): The user can provide the name of the date
            column. If the user-specified column does not exist, all data will be
            retrieved from the table. Defaults to None.
        start_date (str, optional): Parameter to pass start date e.g.
            "2022-01-01". Defaults to None.
        end_date (str, optional): Parameter to pass end date e.g.
            "2022-01-01". Defaults to None.
        columns (list[str], optional): List of columns from given table name.
            Defaults to None.
        adls_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        adls_azure_key_vault_secret (str, optional): The name of the Azure Key Vault
            secret containing a dictionary with ACCOUNT_NAME and Service Principal
            credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
            Defaults to None.
        adls_path (str, optional): Azure Data Lake destination file path.
            Defaults to None.
        adls_path_overwrite (bool, optional): Whether to overwrite the file in ADLS.
            Defaults to True.
    """
    data_frame = bigquery_to_df(
        config_key=config_key,
        azure_key_vault_secret=azure_key_vault_secret,
        query=query,
        dataset_name=dataset_name,
        table_name=table_name,
        date_column_name=date_column_name,
        start_date=start_date,
        end_date=end_date,
        columns=columns,
    )

    return df_to_adls(
        df=data_frame,
        path=adls_path,
        credentials_secret=adls_azure_key_vault_secret,
        config_key=adls_config_key,
        overwrite=adls_path_overwrite,
    )



"""Tasks for interacting with Eurostat."""

import pandas as pd
from prefect import task

from viadot.sources import Eurostat


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def eurostat_to_df(
    dataset_code: str,
    params: dict[str, str] | None = None,
    columns: list[str] | None = None,
    tests: dict | None = None,
) -> pd.DataFrame:
    """Task for creating a pandas DataFrame from Eurostat HTTPS REST API.

    This function serves as an intermediate wrapper between the prefect flow
    and the Eurostat connector:
    - Instantiates an Eurostat Cloud API connector.
    - Creates and returns a pandas DataFrame with the response from the API.

    Args:
        dataset_code (str): The code of the Eurostat dataset that you would like
            to download.
        params (dict[str, str] | None, optional):
            A dictionary with optional URL parameters. Each key is a parameter ID,
            and the value is a specific parameter code, e.g.,
            `params = {'unit': 'EUR'}` where "unit" is the parameter, and "EUR" is
            the code. Only one code per parameter is allowed. Defaults to None.
        columns (list[str] | None, optional):
            A list of column names (as strings) that are required from the dataset.
            Filters the DataFrame to only include the specified columns.
            Defaults to None.
        tests (dict | None, optional):
            A dictionary containing test cases for the data, including:
            - `column_size`: dict{column: size}
            - `column_unique_values`: list[columns]
            - `column_list_to_match`: list[columns]
            - `dataset_row_count`: dict{'min': number, 'max': number}
            - `column_match_regex`: dict{column: 'regex'}
            - `column_sum`: dict{column: {'min': number, 'max': number}}.
            Defaults to None.

    Returns:
        pd.DataFrame:
            A pandas DataFrame containing the data retrieved from the Eurostat API.
    """
    return Eurostat(
        dataset_code=dataset_code, params=params, columns=columns, tests=tests
    ).to_df()



import pendulum
import pytest

from viadot.orchestration.prefect.utils import DynamicDateHandler


DDH1 = DynamicDateHandler(
    ["<<", ">>"], dynamic_date_format="%Y%m%d", dynamic_date_timezone="Europe/Warsaw"
)

DDH2 = DynamicDateHandler(
    ["[[", "]]"], dynamic_date_format="%Y%m%d", dynamic_date_timezone="Europe/Warsaw"
)


@pytest.fixture
def setup_dates():
    """Fixture to provide the current dates for comparison in tests."""
    today = pendulum.today("Europe/Warsaw")
    yesterday = pendulum.yesterday("Europe/Warsaw")
    last_month = today.subtract(months=1).month
    last_year = today.subtract(years=1)
    last_day_prev_month = today.subtract(months=1).end_of("month")
    now_time = pendulum.now("Europe/Warsaw")
    return {
        "today": today.strftime("%Y%m%d"),
        "yesterday": yesterday.strftime("%Y%m%d"),
        "current_month": today.strftime("%m"),
        "last_month": f"{last_month:02d}",
        "current_year": today.strftime("%Y"),
        "last_year": last_year.strftime("%Y"),
        "last_day_previous_month": last_day_prev_month.strftime("%Y%m%d"),
        "now_time": now_time.strftime("%H%M%S"),
    }


def test_create_dates(setup_dates):
    """Test if create_dates function returns correct dictionary values."""
    keys_to_compare = [
        "today",
        "yesterday",
        "current_month",
        "last_month",
        "current_year",
        "last_year",
        "last_day_previous_month",
    ]
    assert all(setup_dates[key] == DDH1.replacements[key] for key in keys_to_compare)


def test_process_dates_single(setup_dates):
    """Test if process_dates function replaces a single date placeholder."""
    text = "Today's date is <<today>>."
    replaced_text = DDH1.process_dates(text)
    expected_text = f"Today's date is {setup_dates['today']}."
    assert replaced_text == expected_text


def test_process_dates_multiple(setup_dates):
    """Test if process_dates function replaces multiple placeholders."""
    text = "Today is <<today>>, yesterday was <<yesterday>>"
    replaced_text = DDH1.process_dates(text)
    expected_text = (
        f"Today is {setup_dates['today']}, yesterday was {setup_dates['yesterday']}"
    )
    assert replaced_text == expected_text


def test_process_dates_with_custom_symbols(setup_dates):
    """Test if process_dates function works with custom start and end symbols."""
    text = "The year is [[current_year]]."
    replaced_text = DDH2.process_dates(text)
    expected_text = f"The year is {setup_dates['current_year']}."
    assert replaced_text == expected_text


def test_process_dates_with_malformed_keys():
    """Test if process_dates function leaves malformed placeholders unchanged."""
    text = "This is a malformed placeholder: <<today."
    replaced_text = DDH1.process_dates(text)
    assert replaced_text == text


def test_process_eval_date_success(setup_dates):
    """Test if process_dates function works with a pendulum code."""
    text = "Yesterday was <<pendulum.today().subtract(days=1)>>."
    replaced_text = DDH1.process_dates(text)
    expected_text = f"Yesterday was {setup_dates['yesterday']}."
    assert replaced_text == expected_text


def test_process_eval_date_fail():
    """Test if process_dates function works with a pendulum code."""
    text = "Yesterday was <<(pendulum.today().subtract(days=1))>>."  # It should not start with `(`
    with pytest.raises(TypeError):
        DDH1.process_dates(text)

    text2 = "Yesterday was <<some_operation=['1','2']>>."  # only `pendulum` works
    with pytest.raises(TypeError):
        DDH1.process_dates(text2)


def test_process_last_x_years():
    """Test if process_dates function returns a date range of last x years."""
    x = 5
    text = f"<<last_{x}_years>>"
    processed_range = DDH1.process_dates(text)
    x_years_ago = pendulum.today().year - (x - 1)
    expected_result = [str(x_years_ago + i) for i in range(x)]
    assert processed_range == expected_result


def test_process_last_x_months():
    """Test if process_dates function returns a date range of last x years."""
    x = 18
    text = f"<<last_{x}_months>>"
    processed_range = DDH1.process_dates(text)
    start_date = pendulum.today().subtract(months=(x - 1))
    expected_result = [
        start_date.add(months=i).start_of("month").format("YMM") for i in range(x)
    ]
    assert processed_range == expected_result


def test_process_years_from_x_until_y_years_ago():
    """Test if `process_dates()` returns a range of years from a given start year."""
    years_ago_count = 3
    start_year = 2019
    text = f"<<years_from_{start_year}_until_{years_ago_count}_years_ago>>"

    processed_range = DDH1.process_dates(text)
    expected_result = [
        str(year)
        for year in range(start_year, pendulum.now().year - years_ago_count + 1)
    ]

    assert processed_range == expected_result


def test_process_years_from_x_until_now():
    """Test if `process_dates()` returns a range of years from a given start year."""
    start_year = 2019
    text = f"<<years_from_{start_year}_until_now>>"

    processed_range = DDH1.process_dates(text)
    expected_result = [str(year) for year in range(start_year, pendulum.now().year + 1)]

    assert processed_range == expected_result


def test_process_string():
    """Test the `_process_string` function to verify it correctly processes dates."""
    result = DDH1._process_string("Let's meet <<today>>.")
    assert result == f"Let's meet {pendulum.today().strftime('%Y%m%d')}."

    result = DDH1._process_string(
        "The event is scheduled for <<pendulum.tomorrow().strftime('%Y-%m-%d')>>."
    )
    assert (
        result
        == f"The event is scheduled for {pendulum.tomorrow().strftime('%Y-%m-%d')}."
    )


def test_find_dynamic_date_patterns():
    """Test the `_find_dynamic_date_patterns` function."""
    query = (
        "SELECT * FROM table"
        " WHERE date_column in (<<today>>, <<yesterday>>, <<2_days_ago>>)"
    )
    result = DDH1._find_dynamic_date_patterns(query)
    expected_result = [
        "<<today>>",
        "<<yesterday>>",
        "<<2_days_ago>>",
    ]
    assert result == expected_result



"""Flow for downloading data from SQLServer and saveing it to a Parquet file."""

from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import sql_server_to_df
from viadot.orchestration.prefect.tasks.task_utils import df_to_parquet


@flow(
    name="extract--sql_server--parquet",
    description="Extract data from SQLServer and save it to a Parquet file.",
    retries=1,
    retry_delay_seconds=60,
)
def sql_server_to_parquet(
    query: str,
    path: str,
    if_exists: Literal["append", "replace", "skip"] = "replace",
    sql_server_credentials_secret: str | None = None,
    sql_server_config_key: str | None = None,
) -> None:
    """Download a file from SQLServer and save it to a Parquet file.

    Args:
        query (str, required): The query to execute on the SQL Server database.
            If the qery doesn't start with "SELECT" returns an empty DataFrame.
        path (str): Path where to save a Parquet file which will be created while
            executing flow.
        if_exists (Literal, optional): What to do if Parquet exists.
            Defaults to "replace".
        sql_server_credentials_secret (str, optional): The name of the secret storing
            the credentialsto the SQLServer. Defaults to None.
            More info on: https://docs.prefect.io/concepts/blocks/
        sql_server_config_key (str, optional): The key in the viadot config
            holding relevant credentials to the SQLServer. Defaults to None.
    """
    df = sql_server_to_df(
        query=query,
        config_key=sql_server_config_key,
        credentials_secret=sql_server_credentials_secret,
    )

    return df_to_parquet(
        df=df,
        path=path,
        if_exists=if_exists,
    )



"""Sharepoint API connector."""

from http import HTTPStatus
import io
from pathlib import Path
import re
from typing import Any, Literal
from urllib.parse import urlparse

import pandas as pd
from pandas._libs.parsers import STR_NA_VALUES
from pydantic import BaseModel, root_validator
import requests
import sharepy
from sharepy.errors import AuthError

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.signals import SKIP
from viadot.sources.base import Source
from viadot.utils import (
    add_viadot_metadata_columns,
    cleanup_df,
    validate,
    validate_and_reorder_dfs_columns,
)


class SharepointCredentials(BaseModel):
    site: str  # Path to sharepoint website (e.g : {tenant_name}.sharepoint.com)
    username: str  # Sharepoint username (e.g username@{tenant_name}.com)
    password: str  # Sharepoint password

    @root_validator(pre=True)
    def is_configured(cls, credentials: dict):  # noqa: N805, ANN201, D102
        site = credentials.get("site")
        username = credentials.get("username")
        password = credentials.get("password")

        if not (site and username and password):
            msg = "'site', 'username', and 'password' credentials are required."
            raise CredentialError(msg)
        return credentials


class Sharepoint(Source):
    DEFAULT_NA_VALUES = tuple(STR_NA_VALUES)

    def __init__(
        self,
        credentials: SharepointCredentials = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Download Excel files from Sharepoint.

        Args:
        credentials (SharepointCredentials): Sharepoint credentials.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials.
        """
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = dict(SharepointCredentials(**raw_creds))
        super().__init__(*args, credentials=validated_creds, **kwargs)

    def get_connection(self) -> sharepy.session.SharePointSession:
        """Establishe a connection to SharePoint.

        Returns:
            sharepy.session.SharePointSession: A session object representing
                the authenticated connection.

        Raises:
            CredentialError: If authentication to SharePoint fails due to incorrect
                credentials.
        """
        try:
            connection = sharepy.connect(
                site=self.credentials.get("site"),
                username=self.credentials.get("username"),
                password=self.credentials.get("password"),
            )
        except AuthError as e:
            site = self.credentials.get("site")
            msg = f"Could not authenticate to {site} with provided credentials."
            raise CredentialError(msg) from e
        return connection

    def download_file(self, url: str, to_path: list | str) -> None:
        """Download a file from Sharepoint to specific location.

        Args:
            url (str): The URL of the file to be downloaded.
            to_path (str): Where to download the file.

        Example:
            download_file(
                url="https://{tenant_name}.sharepoint.com/sites/{directory}/Shared%20Documents/Dashboard/file",
                to_path="file.xlsx"
            )
        """
        conn = self.get_connection()
        conn.getfile(
            url=url,
            filename=to_path,
        )
        conn.close()

    def scan_sharepoint_folder(self, url: str) -> list[str]:
        """Scan Sharepoint folder to get all file URLs of all files within it.

        Args:
            url (str): The URL of the folder to scan.

        Raises:
            ValueError: If the provided URL does not contain the expected '/sites/'
                segment.

        Returns:
            list[str]: List of URLs pointing to each file within the specified
                SharePoint folder.
        """
        conn = self.get_connection()

        parsed_url = urlparse(url)
        path_parts = parsed_url.path.split("/")
        if "sites" in path_parts:
            site_index = (
                path_parts.index("sites") + 2
            )  # +2 to include 'sites' and the next segment
            site_url = f"{parsed_url.scheme}://{parsed_url.netloc}{'/'.join(path_parts[:site_index])}"
            library = "/".join(path_parts[site_index:])
        else:
            message = "URL does not contain '/sites/' segment."
            raise ValueError(message)

        # -> site_url = company.sharepoint.com/sites/site_name/
        # -> library = /shared_documents/folder/sub_folder/final_folder
        endpoint = (
            f"{site_url}/_api/web/GetFolderByServerRelativeUrl('{library}')/Files"
        )
        response = conn.get(endpoint)
        files = response.json().get("d", {}).get("results", [])

        return [f"{site_url}/{library}{file['Name']}" for file in files]

    def _get_file_extension(self, url: str) -> str:
        """Extracts the file extension from a given URL.

        Parameters:
            url (str): The URL from which to extract the file extension.

        Returns:
            str: The file extension, including the leading dot (e.g., '.xlsx').
        """
        # Parse the URL to get the path
        parsed_url = urlparse(url)
        return Path(parsed_url.path).suffix

    def _download_file_stream(self, url: str, **kwargs) -> pd.ExcelFile:
        """Download the contents of a file from SharePoint.

        Returns the data as an in-memory byte stream.

        Args:
            url (str): The URL of the file to download.

        Returns:
            io.BytesIO: An in-memory byte stream containing the file content.
        """
        if "nrows" in kwargs:
            msg = "Parameter 'nrows' is not supported."
            raise ValueError(msg)

        conn = self.get_connection()

        self.logger.info(f"Downloading data from {url}...")
        try:
            response = conn.get(url)
            response.raise_for_status()  # Raise an exception for HTTP errors
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == HTTPStatus.FORBIDDEN:
                self.logger.exception(f"Access denied to file: {url}")
            else:
                self.logger.exception(f"HTTP error when accessing {url}")
            raise
        except Exception:
            self.logger.exception(f"Failed to download file: {url}")
            raise

        bytes_stream = io.BytesIO(response.content)

        try:
            return pd.ExcelFile(bytes_stream)
        except ValueError:
            self.logger.exception(f"Invalid Excel file: {url}")
            raise

    def _is_file(self, url: str) -> bool:
        """Determines whether a provided URL points to a file based on its structure.

        This function uses a regular expression to check if the URL ends with a
        common file extension. It does not make any network requests and purely
        relies on the URL structure for its determination.

        Parameters:
        url (str): The URL to be checked.

        Returns:
        bool: True if the URL is identified as a file based on its extension,
            False otherwise.

        Example:
        >>> _is_file("https://example.com/file.xlsx")
        True
        >>> _is_file("https://example.com/folder/")
        False
        >>> _is_file("https://example.com/folder")
        False
        """
        # Regular expression for matching file extensions
        file_extension_pattern = re.compile(r"\.[a-zA-Z0-9]+$")

        return bool(file_extension_pattern.search(url))

    def _handle_multiple_files(
        self,
        url: str,
        file_sheet_mapping: dict,
        na_values: list[str] | None = None,
        **kwargs,
    ):
        """Handle downloading and parsing multiple Excel files from a SharePoint folder.

        Args:
            url (str): The base URL of the SharePoint folder containing the files.
            file_sheet_mapping (dict): A dictionary mapping file names to sheet names
                or indexes. The keys are file names, and the values are sheet
                names/indices.
            na_values (Optional[list[str]]): Additional strings to recognize as NA/NaN.

        Returns:
            pd.DataFrame: A concatenated DataFrame containing the data from all
                specified files and sheets.

        Raises:
            ValueError: If the file extension is not supported.
        """
        dfs = []
        for file, sheet in file_sheet_mapping.items():
            file_url = url + file
            try:
                df = self._load_and_parse(
                    file_url=file_url, sheet_name=sheet, na_values=na_values, **kwargs
                )
                dfs.append(df)
            except Exception:
                self.logger.exception(f"Failed to load file: {file_url}")
                continue
        if not dfs:
            self.logger.warning("No valid Excel files were loaded.")
            return pd.DataFrame()
        return pd.concat(validate_and_reorder_dfs_columns(dfs))

    def _load_and_parse(
        self,
        file_url: str,
        sheet_name: str | list[str] | None = None,
        na_values: list[str] | None = None,
        **kwargs,
    ):
        """Loads and parses an Excel file from a URL.

        Args:
            file_url (str): The URL of the file to download and parse.
            sheet_name (Optional[Union[str, list[str]]]): The name(s) or index(es) of
                the sheet(s) to parse. If None, all sheets are parsed.
            na_values (Optional[list[str]]): Additional strings to recognize as NA/NaN.
            **kwargs: Additional keyword arguments to pass to the pandas read function.

        Returns:
            pd.DataFrame: The parsed data as a pandas DataFrame.

        Raises:
            ValueError: If the file extension is not supported.
        """
        file_extension = self._get_file_extension(file_url)
        if file_extension not in [".xlsx", ".xlsm", ".xls"]:
            self.logger.error(
                f"Unsupported file extension: {file_extension} for file: {file_url}"
            )
            msg = (
                "Only Excel (.xlsx, .xlsm, .xls) files can be loaded into a DataFrame."
            )
            raise ValueError(msg)
        file_stream = self._download_file_stream(file_url)
        return self._parse_excel(file_stream, sheet_name, na_values, **kwargs)

    def _parse_excel(
        self,
        excel_file: pd.ExcelFile,
        sheet_name: str | list[str] | None = None,
        na_values: list[str] | None = None,
        **kwargs,
    ):
        """Parses an Excel file into a DataFrame. Casts all columns to string.

        Args:
            excel_file: An ExcelFile object containing the data to parse.
            sheet_name (Optional[Union[str, list[str]]]): The name(s) or index(es) of
                the sheet(s) to parse. If None, all sheets are parsed.
            na_values (Optional[list[str]]): Additional strings to recognize as NA/NaN.
            **kwargs: Additional keyword arguments to pass to the pandas read function.

        Returns:
            pd.DataFrame: The parsed data as a pandas DataFrame.
        """
        return pd.concat(
            [
                excel_file.parse(
                    sheet,
                    keep_default_na=False,
                    na_values=na_values or list(self.DEFAULT_NA_VALUES),
                    dtype=str,  # Ensure all columns are read as strings
                    **kwargs,
                )
                for sheet in ([sheet_name] if sheet_name else excel_file.sheet_names)
            ]
        )

    @add_viadot_metadata_columns
    def to_df(
        self,
        url: str,
        sheet_name: str | list[str] | None = None,
        if_empty: Literal["warn", "skip", "fail"] = "warn",
        tests: dict[str, Any] | None = None,
        file_sheet_mapping: dict[str, str | int | list[str]] | None = None,
        na_values: list[str] | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Load an Excel file or files from a SharePoint URL into a pandas DataFrame.

        This method handles downloading the file(s), parsing the content, and converting
        it into a pandas DataFrame. It supports both single file URLs and folder URLs
        with multiple files.

        Args:
            url (str): The URL of the file to be downloaded.
            sheet_name (Optional[Union[str, list, int]], optional): Strings are used for
                sheet names. Integers are used in zero-indexed sheet positions (chart
                sheets do not count as a sheet position). Lists of strings/integers are
                used to request multiple sheets. Specify None to get all worksheets.
                Defaults to None.
            if_empty (Literal["warn", "skip", "fail"], optional): Action to take if
                the DataFrame is empty.
                - "warn": Logs a warning.
                - "skip": Skips the operation.
                - "fail": Raises an error.
                Defaults to "warn".
            tests (Dict[str, Any], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.
            file_sheet_mapping (dict[str, Union[str, int, list[str]]], optional):
                Mapping of file names to sheet names or indices. The keys are file names
                and the values are sheet names/indices. Used when multiple files are
                involved. Defaults to None.
            na_values (list[str], optional): Additional strings to recognize as NA/NaN.
                If list passed, the specific NA values for each column will be
                recognized. Defaults to None.
            kwargs (dict[str, Any], optional): Keyword arguments to pass to
                pd.ExcelFile.parse(). Note that `nrows` is not supported.

        Returns:
            pd.DataFrame: The resulting data as a pandas DataFrame.

        Raises:
            ValueError: If the file extension is not supported or if `if_empty` is set
                to "fail" and the DataFrame is empty.
            SKIP: If `if_empty` is set to "skip" and the DataFrame is empty.
        """
        if self._is_file(url):
            df = self._load_and_parse(
                file_url=url, sheet_name=sheet_name, na_values=na_values, **kwargs
            )
        elif file_sheet_mapping:
            df = self._handle_multiple_files(
                url=url,
                file_sheet_mapping=file_sheet_mapping,
                na_values=na_values,
                **kwargs,
            )
        else:
            list_of_urls = self.scan_sharepoint_folder(url)
            excel_urls = [
                file_url
                for file_url in list_of_urls
                if self._get_file_extension(file_url) == ".xlsx"
            ]
            if not excel_urls:
                self.logger.warning(f"No Excel files found in folder: {url}")
                return pd.DataFrame()
            dfs = [
                self._load_and_parse(
                    file_url=file_url,
                    sheet_name=sheet_name,
                    na_values=na_values,
                    **kwargs,
                )
                for file_url in excel_urls
            ]
            df = pd.concat(validate_and_reorder_dfs_columns(dfs))

        if df.empty:
            try:
                self._handle_if_empty(if_empty)
            except SKIP:
                return pd.DataFrame()
        else:
            self.logger.info(f"Successfully downloaded {len(df)} rows of data.")

        df_clean = cleanup_df(df)

        if tests:
            validate(df=df_clean, tests=tests)

        return df_clean


class SharepointList(Sharepoint):
    """A class to connect to SharePoint lists and retrieve data."""

    def __init__(
        self,
        default_protocol: str | None = "https://",
        credentials: SharepointCredentials = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Initialize the SharepointList connector.

        Args:
            default_protocol (str, optional): The default protocol to use for
                SharePoint URLs.Defaults to "https://".
            credentials (SharepointCredentials, optional): SharePoint credentials.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials.
        """
        self.default_protocol = default_protocol
        super().__init__(
            *args, credentials=credentials, config_key=config_key, **kwargs
        )

    def _find_and_rename_case_insensitive_duplicated_column_names(
        self, df: pd.DataFrame
    ) -> dict:
        """Identifies case-insensitive duplicate column names in a DataFrame.

        This function is necessary because SharePoint lists can have columns
        with the same name but different cases (e.g., "ID" and "Id"),which can cause
        issues when processing the data. It renames these columns by appending a count
        suffix to ensure uniqueness.

        Columns are renamed based on appearance count. For example, if columns include
        ["ID", "Test", "Description", "Id"], the function will create rename mappings
        {"ID": "id_1", "Id": "id_2"} to ensure uniqueness.

        Args:
            df (pd.DataFrame): The input DataFrame.

        Returns:
            dict: A dictionary mapping duplicate column names to their new names.

        Raises:
            TypeError: If input is not a pandas DataFrame.

        Notes:
            This function iterates through the DataFrame's columns, tracking
            case-insensitive duplicates.
            Duplicate columns are renamed by appending a count
            suffix (e.g., "col_1", "col_2").
        """
        columns = df.columns.tolist()
        seen = {}
        rename_dict = {}

        for col in columns:
            col_lower = col.lower()
            if col_lower in seen:
                seen[col_lower] += 1
                rename_dict[col] = f"{col_lower}_{seen[col_lower]}"
            else:
                seen[col_lower] = 1
                # Check if this column needs to be renamed due to future duplicates
                duplicates = [c for c in columns if c.lower() == col_lower]
                if len(duplicates) > 1:
                    rename_dict[col] = f"{col_lower}_1"

        return rename_dict

    def _build_sharepoint_endpoint(
        self, site_url: str, list_site: str, list_name: str
    ) -> str:
        """Build the SharePoint REST API endpoint URL.

        Args:
            site_url: The base SharePoint site URL with protocol
            list_site: The specific site where the list is stored
            list_name: The name of the SharePoint list

        Returns:
            str: The constructed endpoint URL
        """
        return f"{site_url}/sites/{list_site}/_api/web/lists/GetByTitle('{list_name}')/items"

    def _ensure_protocol(self, site_url: str) -> str:
        """Ensure the site URL has the correct protocol.

        Args:
            site_url: The site URL to check

        Returns:
            str: The site URL with the default protocol if needed
        """
        if not site_url.lower().startswith(self.default_protocol.lower()):
            return f"{self.default_protocol}{site_url}"
        return site_url

    def _get_records(
        self, url: str, params: dict | None = None
    ) -> tuple[list[dict], str | None]:
        """Make a request to the SharePoint API and handle common errors.

        Args:
            url: The API endpoint URL
            params: Optional query parameters

        Returns:
            tuple: (data_items, next_link) where data_items is a list of items
                  and next_link is the URL for the next page or None
        """
        conn = self.get_connection()

        try:
            response = conn.get(url, params=params)
            response.raise_for_status()
        except TimeoutError as e:
            msg = f"Request to SharePoint list timed out: {e!s}"
            raise ValueError(msg) from e
        except requests.exceptions.HTTPError as e:
            status_code = getattr(
                getattr(e, "response", None), "status_code", "unknown"
            )
            msg = f"HTTP error {status_code} when retrieving data from SharePoint list"
            raise ValueError(msg) from e
        except Exception as e:
            msg = f"Failed to retrieve data from SharePoint list: {e!s}"
            raise ValueError(msg) from e

        data = response.json().get("d", {})
        items = data.get("results", [])

        # Get the URL for the next page
        next_link = data.get("__next")
        if isinstance(next_link, dict) and "uri" in next_link:
            next_link = next_link["uri"]

        return items, next_link

    def _paginate_list_data(
        self, initial_url: str, params: dict | None = None, list_name: str = ""
    ) -> list[dict]:
        """Handle pagination for SharePoint list data.

        Args:
            initial_url: The initial API endpoint URL
            params: Optional query parameters for the first request
            list_name: Name of the list for error messages

        Returns:
            list: All items from all pages
        """
        all_results = []
        next_url = initial_url
        first_request = True

        while next_url:
            # For first request, include original parameters
            # For subsequent requests, parameters are in the next_url
            current_params = params if first_request else None

            try:
                items, next_url = self._get_records(next_url, current_params)
                all_results.extend(items)
                first_request = False
            except ValueError as e:
                # Add the list_name to the error message
                msg = str(e).replace(
                    "SharePoint list", f"SharePoint list '{list_name}'"
                )
                raise ValueError(msg) from e

        if not all_results:
            msg = f"No items found in SharePoint list {list_name}"
            raise ValueError(msg)

        return all_results

    @add_viadot_metadata_columns
    def to_df(
        self,
        list_name: str,
        list_site: str,
        query: str | None = None,
        select: list[str] | None = None,
        tests: dict | None = None,
    ) -> pd.DataFrame:
        """Retrieve data from a SharePoint list as a pandas DataFrame.

        Args:
            list_site (str): The Sharepoint site on which the list is stored.
            list_name (str): The name of the SharePoint list.
            query (str, optional): A query to filter items. Defaults to None.
            select (list[str], optional): Fields to include in the response.
                Defaults to None.
            tests (Dict[str], optional): A dictionary with optional list of tests
                to verify the output dataframe. If defined, triggers the `validate`
                function from utils. Defaults to None.

        Returns:
            pd.DataFrame: The list data as a DataFrame.

        Raises:
            ValueError: If the list does not exist or the request fails.
        """
        conn = self.get_connection()
        site_url = self._ensure_protocol(conn.site)
        endpoint = self._build_sharepoint_endpoint(site_url, list_site, list_name)

        # Build request parameters
        params = {}
        if query:
            params["$filter"] = query
        if select:
            params["$select"] = ",".join(select)

        # Get all items with pagination handling
        all_results = self._paginate_list_data(endpoint, params, list_name)

        # Convert to DataFrame
        df = pd.DataFrame(all_results)

        # Handle case-insensitive duplicate column names
        rename_dict = self._find_and_rename_case_insensitive_duplicated_column_names(df)
        df = df.rename(columns=rename_dict)

        return validate(df=df, tests=tests) if tests else df



"""'bigquery.py'."""

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import BigQuery


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def bigquery_to_df(
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    query: str | None = None,
    dataset_name: str | None = None,
    table_name: str | None = None,
    date_column_name: str | None = None,
    start_date: str | None = None,
    end_date: str | None = None,
    columns: list[str] | None = None,
) -> pd.DataFrame:
    """Task to download data from BigQuery API.

    Args:
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        azure_key_vault_secret (str, optional): The name of the Azure Key Vault secret
            where credentials are stored. Defaults to None.
        query (str): SQL query to querying data in BigQuery. Format of basic query:
            (SELECT * FROM `{project}.{dataset_name}.{table_name}`). Defaults to None.
        dataset_name (str, optional): Dataset name. Defaults to None.
        table_name (str, optional): Table name. Defaults to None.
        date_column_name (str, optional): The user can provide the name of the date. If
            the user-specified column does not exist, all data will be retrieved from
            the table. Defaults to None.
        start_date (str, optional): Parameter to pass start date e.g.
            "2022-01-01". Defaults to None.
        end_date (str, optional): Parameter to pass end date e.g.
            "2022-01-01". Defaults to None.
        columns (list[str], optional): List of columns from given table name.
                Defaults to None.

    Raises:
        MissingSourceCredentialsError: Credentials were not loaded.

    Returns:
        pd.DataFrame: The response data as a Pandas Data Frame.
    """
    if not (azure_key_vault_secret or config_key):
        raise MissingSourceCredentialsError

    if not config_key:
        credentials = get_credentials(azure_key_vault_secret)

    bigquery = BigQuery(credentials=credentials, config_key=config_key)

    return bigquery.to_df(
        query=query,
        dataset_name=dataset_name,
        table_name=table_name,
        date_column_name=date_column_name,
        start_date=start_date,
        end_date=end_date,
        columns=columns,
    )
